/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// Code notes:
//
// READ THE HEADER OF THE 'addresses.h' FILE ON WHY IT IS DONE THIS WAY.
//
// sample host<->address table can be defined in your code via:
//     struct name_mapping_raw map_addrs_raw[] = HOST_ADDRESSES_TABLE;
//     struct name_mapping     *map_addrs;
// and then
//     map_addrs = process_mapping(map_addrs_raw);
//
// See the 'killer' program as an example.
//
// Most of these functions are only called once or twice,
//   so we don't really need to worry about maximal efficiency.
// #############################################################################

#include "addresses.h"
#include "kernel_user.h"

#ifndef __KERNEL__
#include <stdio.h>  // for NULL
#include <string.h> // for memset
#include <stdlib.h> // malloc
#else
#include <linux/modversions.h> 
#include <linux/module.h>
#include <linux/tty.h>
#include <linux/slab.h> // kmalloc, kfree, etc.
//OLD#include <linux/kernel.h> // for printk... but it doesn't use modversions????
//OLD#include <asm/string.h> // memset
#endif

// =============================================================================
// dump a binary buffer in hex.
// =============================================================================
void bindump(void *buf, int len) {
	int idx;
	unsigned char *ch=(unsigned char *)buf;

	for (idx=0; idx<len; idx++) {
		PRINT_FUNCTION(PRIORITY_NORMAL "%02X ", ch[idx]);
		if ((idx+1)%16==0) {
			PRINT_FUNCTION("\n");
		}
	}
	PRINT_FUNCTION("\n");
}


// =============================================================================
// dumb, dumb, dumb.
// *stupid* method to convert hex character to deximal number.
// I'm sure there's a library function or something that I'm forgetting.
// =============================================================================
int hextodec(char c) {
	if (c>='a' && c<='f') {
		return(c-'a'+10);
	} else if (c>='A' && c<='F') {
		return(c-'A'+10);
	} else if (c>='0' && c<='9') {
		return(c-'0');
	} else {
		PRINT_FUNCTION(PRIORITY_ERROR "Bad parse (%c,%d).\n", c, (int)c);
		return(-1);
	}
}


// =============================================================================
// Convert a MAC address into a string, assumes string long enough (18 chars).
// =============================================================================
char *mac2str(const unsigned char macaddr[ETH_ALEN], char *macstr) {
	int idx;
	char *p=macstr;

	for (idx=0; idx<(ETH_ALEN-1); idx++) {
		p += sprintf(p, "%02X:", macaddr[idx]);
	}
	sprintf(p, "%02X", macaddr[idx]);

	return macstr;
}


// =============================================================================
// Convert an IP address in host (x86, little-endian order) into a string.
// =============================================================================
char *ip2str_hbo(const unsigned int ipaddr, unsigned char *ipstr) {
	sprintf(ipstr, "%03d.%03d.%03d.%03d",
			(ipaddr&0xFF000000)>>24,
			(ipaddr&0xFF0000)>>16,
			(ipaddr&0xFF00)>>8,
			(ipaddr&0xFF));
	return ipstr;
}


// =============================================================================
// Convert an IP address in network byte order (big-endian) into a string.
// =============================================================================
char *ip2str_nbo(const unsigned int ipaddr, unsigned char *ipstr) {
	sprintf(ipstr, "%03d.%03d.%03d.%03d",
			(ipaddr&0xFF),
			(ipaddr&0xFF00)>>8,
			(ipaddr&0xFF0000)>>16,
			(ipaddr&0xFF000000)>>24);
	return ipstr;
}


// =============================================================================
// given a MAC address string, build the MAC from it.
// =============================================================================
int str2mac(char *str, unsigned char mac[ETH_ALEN]) {
	int idx, jdx;

	if (str==NULL || strlen(str)!=(3*ETH_ALEN-1)) {
		return 0;
	}

	// Error checking on input added; 20030109.
	for (idx=0; idx<ETH_ALEN; idx++) {
		if ((jdx = hextodec(str[idx*3])*16)<0) { return 0; }
		mac[idx] = jdx;
		if ((jdx = hextodec(str[idx*3+1]))<0) { return 0; }
		mac[idx] += jdx;
	}

	return 1;
}


// =============================================================================
// given an ip address string, return an NETWORK B.O. ip integer.
// =============================================================================
unsigned int str2ip(char *str, unsigned int *ip) {
	unsigned int loc_ip=0;
	unsigned char *oct[4];
	int octet, digit;

	if (str==NULL || ip==NULL) {
		return 0;
	}

	oct[0]=&(((unsigned char *)&loc_ip)[0]);
	oct[1]=&(((unsigned char *)&loc_ip)[1]);
	oct[2]=&(((unsigned char *)&loc_ip)[2]);
	oct[3]=&(((unsigned char *)&loc_ip)[3]);

	// process each octet
	// Hand-parse, want to avoid sscanf();
	for (octet=0; octet<4; octet++) {
		*(oct[octet])=0;
		// process up to three digits of decimal data & a '.'
		for (digit=0; digit<4; digit++) {
			// stop when you hit the '.'
			if (str[digit]=='.') {
				// move to the next octet
				str = str+digit+1;
				break;
			} else if (str[digit]>='0' && str[digit]<='9') {
				// add in the decimal value of that digit
				*(oct[octet]) = (*(oct[octet]))*10 + str[digit]-'0';
			} else if (str[digit]=='\0') {
				// check for errors; otherwise done with parse.
				if (octet<3) {
					*ip =0;
					return 0;
				} else {
					break;
				}
			} else {
				// bad digit hit!
				*ip =0;
				return 0;
			}
		}
	}

	// no errors hit...
	*ip = loc_ip;

	return *ip;
}


// =============================================================================
// Convert the raw strings that are convenient to type into network
// byte order IP and MAC addresses.
// =============================================================================
struct name_mapping *process_mapping(struct name_mapping_raw *rawnames) {
	int incount, outcount;
	struct name_mapping *mapping;
	struct name_mapping *addrinfo;

	// error-check
	if (rawnames==NULL) {
		PRINT_FUNCTION(PRIORITY_ERROR "Bad call to process_mapping (NULL)\n");
		return NULL;
	}

	// get count of elements first.
	for (incount=0; rawnames[incount].name!=NULL; incount++)
		;
	incount++;

	// allocate new array...
	if ((mapping=(struct name_mapping *)ALLOC_FUNCTION(incount*sizeof(struct name_mapping) ALLOC_PRIORITY))==NULL) {
		PRINT_FUNCTION(PRIORITY_ERROR "[k]malloc of the name-mapping array failed!\n");
		return NULL;
	}

	// convert each element of array.
	for (incount=outcount=0; rawnames[incount].name!=NULL; incount++) {
		// name first, simply get copy
		if ((mapping[outcount].name=(char *)ALLOC_FUNCTION(strlen(rawnames[incount].name)+1 ALLOC_PRIORITY))==NULL) {
			PRINT_FUNCTION(PRIORITY_ERROR "[k]malloc of a machine name failed, attempting to continue!\n");
			continue;
		}
		memcpy(mapping[outcount].name, rawnames[incount].name, strlen(rawnames[incount].name)+1);

		// then ip
		if (str2ip(rawnames[incount].ip, &mapping[outcount].ip)==0) {
			PRINT_FUNCTION(PRIORITY_ERROR "Failed to parse IP (%s) for machine '%s'. Removing machine from list.\n", rawnames[incount].ip, mapping[outcount].name);
			FREE_FUNCTION(mapping[outcount].name);
			continue;
		}

		// MAC
		if (str2mac(rawnames[incount].mac, mapping[outcount].mac)<=0) {
			PRINT_FUNCTION(PRIORITY_ERROR "Failed to parse MAC (%s) for machine '%s'. Removing machine from list.\n", rawnames[incount].mac, mapping[outcount].name);
			FREE_FUNCTION(mapping[outcount].name);
			continue;
		}

		// Then Mode. Just copy.
		mapping[outcount].mode = rawnames[incount].mode;

		outcount++;
	}

	// terminate array.
	memset(mapping+outcount, '\0', sizeof(struct name_mapping));

	// lastly, the partners. Convert from name to index into the array.
	// This has to be done after everything else because we use the lookup function to do it.
	// foreach host...
	for (incount=0; mapping[incount].name!=NULL; incount++) {
		// foreach partner
		for (outcount=0; outcount<ROUTING_PARTNERS; outcount++) {
			mapping[incount].partners[outcount]=-1;
			// look it up
			if (rawnames[incount].partners[outcount]!=NULL) {
				addrinfo=get_addr_info(mapping,rawnames[incount].partners[outcount], 0, NULL);
				// put index into output if I found the mapping.
				if (addrinfo!=NULL) {
					mapping[incount].partners[outcount]=addrinfo-mapping;
				} else {
					PRINT_FUNCTION(PRIORITY_ERROR "Failed to match output target %s\n", rawnames[incount].partners[outcount]);
				}
			}
		}
	}

	return mapping;
}


// =============================================================================
// Print a single address mapping
// =============================================================================
void print_one_map(struct name_mapping *map) {
	char tmpbuf1[32];
	char tmpbuf2[32];
	int hit;

	PRINT_FUNCTION(PRIORITY_NORMAL "    %-16s  %s  %s",
			map->name,
			ip2str_nbo(map->ip, tmpbuf1),
			mac2str(map->mac, tmpbuf2));
	hit=0;
	if (map->mode&MODE_INPUT ) { PRINT_FUNCTION(" INPUT ");  hit++; }
	if (map->mode&MODE_OUTPUT) { PRINT_FUNCTION(" OUTPUT"); hit++; }
	if (map->mode&MODE_TARGET) { PRINT_FUNCTION(" TARGET"); hit++; }
	if (!hit) { PRINT_FUNCTION(" NONE"); }

	for (hit=0; hit<ROUTING_PARTNERS; hit++) {
		if (map->partners[hit]>=0) {
			PRINT_FUNCTION(" %d", map->partners[hit]);
		}
	}

	PRINT_FUNCTION("\n");
}


// =============================================================================
// Print known address mappings.
// =============================================================================
void print_mapping(struct name_mapping *map) {
	int idx;

	PRINT_FUNCTION(PRIORITY_NORMAL "    hostname          IP address       MAC address       MODES  partners\n");
	for (idx=0; map[idx].name!=NULL; idx++) {
		print_one_map(map+idx);
	}
}


// =============================================================================
// Destroy the mapping created in 'process_mapping', above.
// Make sure to not use it after calling this function.
// =============================================================================
void free_mapping(struct name_mapping *processednames) {
	int idx;

	if (processednames!=NULL) {
		// first all the strings we created
		for (idx=0; processednames[idx].name!=NULL; idx++) {
			FREE_FUNCTION(processednames[idx].name);
		}

		// last the array itself.
		FREE_FUNCTION(processednames);
		processednames=NULL;
	}
}


// =============================================================================
// given a host name, ip string, or mac address, return all info.
// Assumes only one argument is filled in and rest are null;
//    if not, matches on first nonnull.
// 20020215: Removed print statements; caller must print errors, if wants.
// =============================================================================
struct name_mapping *get_addr_info(struct name_mapping *table, char *name, unsigned int ip, unsigned char *mac) {
	int idx;

	// Ordering this way because IP check is most commonly used.
	if (ip != 0) {
		for (idx=0; table[idx].name; idx++) {
			if (ip==table[idx].ip) {
				// Match! return this record.
				return(&(table[idx]));
			}
		}

		return NULL;
	} else if (name != NULL) {
		// First, unqualify domain name...
		for (idx=0; idx<strlen(name); idx++) {
			if (name[idx]=='.') {
				name[idx] = '\0';
				break;
			}
		}

		// Loop over names/addresses in the array, and see if names match.
		for (idx=0; table[idx].name; idx++) {
			if (strcmp(name, table[idx].name)==0) {
				// match. return this record.
				return(&(table[idx]));
			}
		}

		// debug: PRINT_FUNCTION(PRIORITY_ERROR "Failed to match host name '%s'\n", name);
		return NULL;
	} else if (mac != NULL) {
		// Loop over names/addresses in the array, and see if names match.
		for (idx=0; table[idx].name; idx++) {
			if (memcmp(mac, table[idx].mac, ETH_ALEN)==0) {
				// match. return this record.
				return(&(table[idx]));
			}
		}

		// debug: PRINT_FUNCTION(PRIORITY_ERROR "Failed to match mac address!\n");
		return NULL;
	}

	// debug: PRINT_FUNCTION(KERN_EMERG "3 Nulls passed to get_addr_info. You are dumb.\n");
	return NULL;
}

