/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// Dump header information, shared between user and kernel programs:
//    kernel/dump.c, user{reader, writer, printer}
//
// Gives info on collected information, record sizes, etc.
// #############################################################################

#ifndef TICKET_COMMON_DUMP
#define TICKET_COMMON_DUMP

// ================================================================================
// CONFIGURATION:
// ================================================================================

// --------------------------------------------------------------------------------
// Mode: TCPDUMP drop-in replacement (http://www.tcpdump.org) OR
//       TICKET; just get specific headers...
// --------------------------------------------------------------------------------
//#define DUMP_MODE_TICKET
#define DUMP_MODE_TICKET

#if defined(DUMP_MODE_TICKET) && defined(DUMP_MODE_TCPDUMP)
// bad things will happen to you and your children unto the seventh generation.
#error "ERROR! You must PICK and #define ONE-- TICKET or TCPDUMP dumping!"
#endif

// --------------------------------------------------------------------------------
// Capture configuration. If doing tcpdump capture; how many bytes to save?
// WARNING: must be below 1500-20-8-8 = 1464. I can't give you whole packets.
// The smaller this is, the more efficient collection is. The larger, the more
// the dump code turns into a round-robin packet forwarder.
//
// `man tcpdump` to see why the default is 68.
// --------------------------------------------------------------------------------
#define OUT_SNAPLEN 68

// --------------------------------------------------------------------------------
// Striping mechanism (writer.c)
//
// Split files around 200MB;
//     Larger files will: leave larger spaces (and hence, wasted space)
//        at the end of disks given the striping logic.
//     Smaller files will: cause more overhead per file,
//        possibly leading to loss of packets.
// Pick your poison.
// --------------------------------------------------------------------------------
// How many packets to save per file, before splitting?
#define RECORDS_PER_FILE ((unsigned)((200*1024*1024)/sizeof(dump_packet)))
// how many user bytes per file?
#define FILE_SIZE_BYTES  (RECORDS_PER_FILE*sizeof(dump_packet))
// how many blocks per file? Add 4k for directory structure and rounding.
#define FILE_SIZE_BLOCKS ((unsigned)((RECORDS_PER_FILE*sizeof(dump_packet))/1024) + 4)
// How many files to put in each directory.
#define DUMP_FILES_PER_DIR 100


// =============================================================================
// Defines of useful stuff
// =============================================================================

// For header sizes
#include "net_headers.h"

// ABSOLUTE dump header length
// should be sizeof(unsigned long long)
#define DUMP_HEADER_LEN 8

// Maximum size of packet collected off the wire
#define IN_SNAPLEN (ETH_HEADER_LEN+ETH_DATA_LEN)

// how big our dump record is.
#define DUMP_RECORD_LEN (sizeof(collected_headers))

// Total overhead of headers in a dump packet.
#define DUMP_OVERHEAD_LEN (ETH_HEADER_LEN+IP_HEADER_LEN+UDP_HEADER_LEN+DUMP_HEADER_LEN)

// Maximum space available to our collected headers.
#define MAX_DUMP_LEN (IN_SNAPLEN-DUMP_OVERHEAD_LEN)

// (1500-overheads/DUMP_RECORD_LEN) = 34 (total 1470+8TS == 22 bytes free)
#define DUMPS_PER_PACKET (MAX_DUMP_LEN/DUMP_RECORD_LEN)

// Amount of space left for alignment, etc.
#define LEFTOVER_SPACE (MAX_DUMP_LEN - (DUMPS_PER_PACKET*DUMP_RECORD_LEN))


// =============================================================================
// The actual stuff we're collecting.
//
// TODO: Version with just 68 byte character array; selected via DUMP_OUTPUT_FORMAT
//       option in config file. (see common/config.h)
// =============================================================================
struct collected_headers_ticket {
	// Basic information.
	unsigned long long timestamp; // cycle counters
	unsigned short length;        // 'off-the-wire' length (in bytes, up to 1514<2048=2^11)

	// Ethernet/802.3
	unsigned char mac_dst[ETH_ALEN]; // ethernet
	unsigned char mac_src[ETH_ALEN]; // ethernet
	unsigned short toslen;    // MAC TOS/802.3 length (usually 0x0800==IP)

	// IP information
	unsigned int ip_src;      // IP source
	unsigned int ip_dst;      // IP destination
	unsigned short ip_len;    // IP packet length
	unsigned char ip_hdr_len; // IP header length
	unsigned char protocol;   // IP protocol field

	// TCP/UDP information
	unsigned short src_port;  // TCP/UDP source port
	unsigned short dst_port;  // TCP/UDP destination port
	union {
		unsigned short udp_len; // UDP length field (in bytes)
		// len must come BEFORE pad to print
		// correctly as short on little-endian box.
		struct {
			unsigned char len;  // TCP header length field (in 32-bit words)
			unsigned char pad;  // empty; ignore!
		} tcp_len;
	} upper_len;
} __attribute__ ((packed));

/* convert pktlen to caplen and len as follows:
 *     if pktlen<=OUT_SNAPLEN then caplen = pktlen,      len=pktlen
 *     if pktlen> OUT_SNAPLEN then caplen = OUT_SNAPLEN, len=pktlen */
struct collected_headers_tcpdump {
	unsigned long long timestamp; // cycle counters
	unsigned int pktlen; /* off-the-wire. */
	unsigned char snap[OUT_SNAPLEN];
} __attribute__ ((packed));

#ifdef DUMP_MODE_TICKET
	typedef struct collected_headers_ticket collected_headers;
#endif
#ifdef DUMP_MODE_TCPDUMP
	typedef struct collected_headers_tcpdump collected_headers;
#endif

struct dump_packet_ticket {
	unsigned long long seqno;
	struct collected_headers_ticket data[DUMPS_PER_PACKET];
};

struct dump_packet_tcpdump {
	unsigned long long seqno;
	struct collected_headers_tcpdump data[DUMPS_PER_PACKET];
};

#ifdef DUMP_MODE_TICKET
	typedef struct dump_packet_ticket dump_packet;
#endif
#ifdef DUMP_MODE_TCPDUMP
	typedef struct dump_packet_tcpdump dump_packet;
#endif

#endif

