/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// Routing mode. use TICKET as a dumb WAN emulator.
// #############################################################################

/* Standard headers for kernel modules */
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/tty.h>

/* network headers, netif_rx */
#include <linux/netdevice.h>
/* for rtnl_shlock(), etc (need to hold to use dev_get_by_index, dev_open) */
#include <linux/rtnetlink.h>
#include <linux/sched.h> /* for doing non-blocking sleeps */
#include <linux/wait.h>

#include "../common/config.h"
#include "../common/net_headers.h"
#include "../common/network.h"
#include "../common/addresses.h"
#include "../common/route.h"
#include "../common/kthread.h"
#include <asm/smplock.h>     /* (un)*lock_kernel */

// configuration. see common/config.h for more on this structure.
struct routing_config {
	unsigned long idx;      // UNUSED. Just for convenience in writing.
	unsigned long delay;    // milliseconds to delay each packet.
	unsigned long duration; // how long (milleseconds) to maintain this delay mode. (SEMANTICS CHANGE WARNING: 0 means core dump!)
	unsigned long drop_n;   // drop every Nth packet. (0 means drop nothing. 1 means drop everything)
	int next_delay;         // 'goto' index in array of next delay when this one expires.
};

// =============================================================================
// File-global variables.
// =============================================================================

struct ew_thread packet_thread; /* ALL thread data */
int pass_through=1; /* DO pass stuff to netif_rx after I see it */
unsigned long ticket_good_drops=0;
unsigned long ticket_evil_drops=0;
int (*old_netif_rx)(struct sk_buff *skb);
#if LINUX_VERSION_CODE>=0x020420
int (*old_netif_receive)(struct sk_buff *skb);
#endif
extern struct name_mapping *map_addrs, *broadcast_addr; /* from network.c */

// The delay characteristics we have. See ./common/config.h
struct routing_config routing_states[] = { ROUTING_CONFIG };

// The queue to delay skbs.
struct sk_buff_head ticket_skbs;

// =============================================================================
// Forward declarations.
// =============================================================================
void ticket_route_exit(void);


// =============================================================================
// Route a packet (socket buffer). Interrupt context (?) Can only use spinlocks (?)
// Generalized to handle new NAPI interface.
// =============================================================================
int netif_rex_route(struct sk_buff *skb, int (*pass_through_function)(struct sk_buff *skb)) {
	struct sk_buff *copy=NULL;
	static int copy_count=10000;

	// debug printk(KERN_EMERG "netif_rex_route handling packet for %p.\n", pass_through_function);

	while(copy_count>0) {
		copy_count--;

		// First, check the packet. It should have the eth. headers removed,
		// but still valid (in the packet).
		if ((skb->data - skb->mac.raw)==ETH_HEADER_LEN) {
			// if passing through, get a copy!
			if (pass_through) {
				copy = skb_copy(skb, GFP_ATOMIC);
			} else {
				copy = skb;
				skb = NULL;
			}

			if (copy) {
				// can only queue so many packets before we're screwed.
				if (ticket_skbs.qlen>=TICKET_ROUTE_BACKLOG) {
					// debug printk(KERN_WARNING "Backlog, dropping.");
					kfree_skb(skb);
					ticket_evil_drops++;
					return NET_RX_DROP;
				}

				// Timestamp as soon as possible.
				// (trying to minimize work done before queueing;
				//    queue length value checks first!)
				// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
				do_gettimeofday(&copy->stamp);

				// Enqueue the packet for later.
				// Gets spinlock, so is 'safe'
				skb_queue_tail(&ticket_skbs, copy);

				// notify init that there's a packet.
				up(&packet_thread.packet_wait_sem);
			} else {
				printk("Can't copy my skb!");
				ticket_evil_drops++;
			}
		} else {
			printk(KERN_WARNING "Nonforwardable packet recieved.");
			ticket_good_drops++;
		}

		// maybe Linux can do something with it, though...
		if (skb) {
			if (pass_through) {
				return(pass_through_function(skb));
			} else {
				kfree_skb(skb);
			}
		}
		return NET_RX_DROP;
	}

	return(pass_through_function(skb));
}


// =============================================================================
// First-order wrappers to map old softnet and new NAPI interfaces to one
// stupid generalized function.
// =============================================================================
int netif_rx_route(struct sk_buff *skb) {
	return(netif_rex_route(skb, old_netif_rx));
}
#if LINUX_VERSION_CODE>=0x020420
int netif_receive_route(struct sk_buff *skb) {
	return(netif_rex_route(skb, old_netif_receive));
}
#endif

// =============================================================================
// get diff in microseconds.
// `later' damn well be after `before'.
// =============================================================================
unsigned long mu_s_diff(struct timeval *before, struct timeval *later) {
	unsigned long diff;

	diff = (1000000 * ((unsigned long)later->tv_sec - (unsigned long)before->tv_sec));
	diff += (unsigned long)later->tv_usec;
	diff -= (unsigned long)before->tv_usec;

	return(diff);
}


// =============================================================================
// return next delay index, tell user what's up.
// TODO: proceed past other states by looping and time subtraction;
//       may need to move past multiple states at once!
// =============================================================================
int change_delay(int cur_delay, struct timeval *tv_start, struct timeval *tv_now, int *drop_idx) {
	// move to next state.
	cur_delay=routing_states[cur_delay].next_delay;

	printk("Delay change. state[%d]: %lu ms, Duration: %ld ms, Drop every %ldth, Next index %d.\n",
			cur_delay,
			routing_states[cur_delay].delay,
			routing_states[cur_delay].duration,
			routing_states[cur_delay].drop_n,
			routing_states[cur_delay].next_delay);

	// book-keeping.
	*drop_idx=0;
	do_gettimeofday(tv_start); // 20021205; Between 2.4.5 & 2.4.20... 'get_fast_time' renamed to 'do_gettimeofday'
	do_gettimeofday(tv_now);   // required to avoid double progression

	// Done! return new delay index.
	return(cur_delay);
}


// =============================================================================
// Main thread routine to do all routing stuff.
// =============================================================================
void ticket_route_thread(void) {
	// The current packet whose data is being collected.
	struct sk_buff *input_skb=NULL;
	struct timeval tv_start, tv_now;
	int drop_idx, idx, cur_delay=0;
	struct name_mapping *dst_addr, *input_map, *output_map, *target_map;
	// debug char tmpbuf1[64], tmpbuf2[64];
	DECLARE_WAIT_QUEUE_HEAD(waitq_local);
	DECLARE_WAITQUEUE(wait, current);

	lock_kernel(); /* set up thread environment; get BKL to modify low-level process stuff. */
	packet_thread.thread = current; /* fill in thread structure */
	sprintf(current->comm, "ticket_route"); /* set name of this process; sizeof(comm)==16 */
	init_waitqueue_head(&waitq_local); /* TODO don't need BKL, do I??? */
	unlock_kernel(); /* let others run */
	up(&packet_thread.control_sem); /* Completed initialization; let creator finish */

	/* Must grab this late in the game; don't want to muck with control
	 * flow by getting packets in interrupt context for thread we're starting. */
	// debug printk(KERN_INFO "changing netif_rx(s) from %p to %p\n", netif_rx, &netif_rx_route);
	netif_rx = &netif_rx_route;
#if LINUX_VERSION_CODE>=0x020420
	netif_receive_skb = &netif_receive_route;
#endif

	// convert delay from ms to us; do this before the print to detect funkiness.
 	// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
	do_gettimeofday(&tv_start);

	printk(KERN_INFO "TICKET router (WAN emulator) thread running; stole netif_rx.\n");
	printk("%d: Delaying packets for %lu ms each way, duration %ld.\n",
			cur_delay, routing_states[cur_delay].delay, routing_states[cur_delay].duration);

	// Set all input devices promiscuous.
	for (input_map=map_addrs; ew_eth[input_map-map_addrs]!=NULL; input_map++) {
		if (input_map->mode & MODE_INPUT) {
			dev_set_promiscuity(ew_eth[input_map-map_addrs], 1);
		}
	}

	// Loop "forever" (break if packet_thread.die, later)
	drop_idx=0;
	while (1) {
		// Put myself to sleep; when wakes, there's a packet to handle.
		down(&packet_thread.packet_wait_sem);

		/* Make sure flags are visible on all CPUs. */
		/* EWEW??? Apparently this can be pretty darn painful (expensive). */
		mb();

		/* Are we back to die, or to process a packet? */
		if (packet_thread.die) {
			break;
		}

		/* this is normal work to do */
		// debug printk("Got a packet\n");

		// Get the packet off the list.
		if ((input_skb = skb_dequeue(&ticket_skbs))==NULL) {
			printk("Dequeued NULL skb in init!!!\n");
			continue; // shouldn't happen, but don't die.
		}

#ifdef ALLOW_INSECURE_ADMIN
			// (Network reboot code. C break-out logic makes it fairly low cost.)
			if (input_skb->len >= (ETH_HEADER_LEN+4) && input_skb->mac.ethernet->h_proto==0x9999) {
				if (memcmp(&(input_skb->data[ETH_HEADER_LEN]), "DIE!", 4)==0) {
					printk(KERN_EMERG "Reboot request received, but not configured yet!\n");
					machine_restart(NULL); // TODO: check
				} else if (memcmp(&(input_skb->data[ETH_HEADER_LEN]), "NEXT", 4)==0) {
					cur_delay = change_delay(cur_delay, &tv_start, &tv_now, &drop_idx);
				} else {
					printk(KERN_EMERG "Unknown administration attempt.\n");
				}
				goto route_drop;
			}
#endif

		drop_idx++;
		// 20021205; this got renamed between 2.4.5 and 2.4.20... get_fast_time --> do_gettimeofday
		do_gettimeofday(&tv_now);

		// We want the headers back!
		input_skb->nh.iph = (struct iphdr *)input_skb->data;
		skb_push(input_skb, ETH_HEADER_LEN);
		if (input_skb->len < ETH_HEADER_LEN) {
			printk("Got a freakishly short ethernet packet!");
			goto route_drop;
		}
		input_skb->mac.ethernet = (struct ethhdr *)input_skb->data;

		// ------------------------------------------------------------
		// Which device did it come in on?
		// TODO: this is kind of slow... Worth it?
		// ------------------------------------------------------------
		input_map = map_addrs;
		while (ew_eth[input_map-map_addrs]!=input_skb->dev && ew_eth[input_map-map_addrs]) {
			 input_map++;
		}
		if (ew_eth[input_map-map_addrs]!=input_skb->dev) {
			// debug printk("Failed to look up input device %s\n", input_skb->dev->name);
			// This shouldn't happen.
			goto route_drop;
		}
		// debug printk("Received input on %s (%s)\n", input_map->name, input_skb->dev->name);
		if (!(input_map->mode & MODE_INPUT)) {
			// debug printk("Ignoring a packet on a non-input interface.\n");
			goto route_drop;
		}

		// ----------------------------------------
		// Packets to drop
		// ----------------------------------------
		if (!memcmp(input_skb->mac.ethernet->h_dest, broadcast_addr->mac, ETH_ALEN)) {
			// debug printk("ENET: Drop broadcast from %s, destination should already get.\n",
			// debug 	mac2str(input_skb->mac.ethernet->h_source, tmpbuf1));
			goto route_drop;
		}
		for (idx=0; idx<ETHERNET_IF_COUNT && ew_eth[idx]; idx++) {
			if (memcmp(ew_eth[idx]->dev_addr, input_skb->mac.ethernet->h_source, ETH_ALEN)==0) {
				// debug printk("Ethernet: Drop looped packet (source MAC is that of device %d)\n", idx);
				goto route_drop;
			}
		}
		if (input_skb->mac.ethernet->h_proto!=ehw_htons(ETH_P_IP) || input_skb->len < (ETH_HEADER_LEN+IP_HEADER_LEN)) {
			// debug printk("Ethernet: Drop non-IP packet (can't route), protocol 0x%04X\n",
			// debug 		ehw_ntohs(input_skb->mac.ethernet->h_proto));
			goto route_drop;
		}
		if (input_skb->nh.iph->daddr==broadcast_addr->ip) {
			// debug printk("IP: Drop broadcast from %s, destination should already get.\n",
			// debug 		ip2str_nbo(input_skb->nh.iph->saddr, tmpbuf1));
			goto route_drop;
		}
		if (!(dst_addr=get_addr_info(map_addrs, NULL, input_skb->nh.iph->daddr, NULL))) {
			// debug printk("IP: Drop un-routable packet (failed to match IP address %s)\n",
			// debug 		ip2str_nbo(input_skb->nh.iph->daddr, tmpbuf1));
			goto route_drop;
		}
		if (routing_states[cur_delay].drop_n && drop_idx==routing_states[cur_delay].drop_n) {
			// debug printk("Drop packets if user says so in routing table... it's time!\n");
			goto route_drop;
		}
		//TODO: drop if it's to localhost as the end point!!!

		// ----------------------------------------
		// Has packet been delayed long enough? (remember, routing_states in MILLISECONDS, timevals in MICROSECONDS)
		// ----------------------------------------
		// debug printk("Stamp: %lu.%lu, now %lu.%lu\n", input_skb->stamp.tv_sec, input_skb->stamp.tv_usec,
		// debug 		tv_now.tv_sec, tv_now.tv_usec);
		// TODO This code assumes HZ=100 so 1/HZ = .01s = 10ms = 10*1000us.
		while (mu_s_diff(&input_skb->stamp, &tv_now)+10000 < (routing_states[cur_delay].delay*1000)) {
			// debug printk("Waiting a long time (%luus)\n",
			// debug 		routing_states[cur_delay].delay*1000-mu_s_diff(&input_skb->stamp, &tv_now));
			add_wait_queue(&waitq_local, &wait); /* what is wait? wait_queue_t *new */
			set_current_state(TASK_INTERRUPTIBLE);
			(void)schedule_timeout(1); /* 1 tick, 10ms. ignore return value; we'll check time below! */
			remove_wait_queue(&waitq_local, &wait);
			do_gettimeofday(&tv_now);
		}

		// short diffs get busy-waits.
		for (do_gettimeofday(&tv_now);
				mu_s_diff(&input_skb->stamp, &tv_now)<(routing_states[cur_delay].delay*1000); 
				do_gettimeofday(&tv_now)) {
			// debug printk("Waiting a short time (%luus)\n",
			// debug 		routing_states[cur_delay].delay*1000-mu_s_diff(&input_skb->stamp, &tv_now));
			// this gives a 1ms delay. Might want to decrease.
			udelay(1000);
		}

		// Get stripe output.
		if (!next_output(&output_map, &target_map)) {
			printk("Failed to find _any_ output device!\n");
			printk("This is a configuration error.\n");
			goto route_drop;
		}

		// put device into skb (recall map_addrs and ew_eth sorted identically; get same idx'd device).
		input_skb->dev = ew_eth[output_map-map_addrs];

		// Put in proper MAC addresses...
		//   source address should become:      OUTGOING INTERFACE
		//   destination address should become: TARGET BOX'S INTERFACE
		memcpy(input_skb->mac.ethernet->h_source, input_skb->dev->dev_addr, input_skb->dev->addr_len);
		memcpy(input_skb->mac.ethernet->h_dest,   dst_addr->mac,      input_skb->dev->addr_len);

		// force the packet to be sent *NOW*. The transmission code frees the input_skb.
		// Don't muck with checksum... Otherwise will re-checksum with wrong pointers and corrupt packet!!!!
		// debug printk(KERN_INFO "Sending to %s, %s, %s via %s.\n",
		// debug 		dst_addr->name,
		// debug 		ip2str_nbo(input_skb->nh.iph->daddr, tmpbuf1),
		// debug 		mac2str(input_skb->mac.ethernet->h_dest, tmpbuf2),
		// debug 		input_skb->dev->name);
		input_skb->ip_summed = CHECKSUM_NONE;
		// 20020912: spin lock should not be required; there is no contention.
		// spin_lock(&input_skb->dev->xmit_lock);
		input_skb->dev->hard_start_xmit(input_skb, input_skb->dev);
		// spin_unlock(&input_skb->dev->xmit_lock);
		input_skb=NULL;

route_drop:
		// If we make it to here with a packet, it _should_ be dropped.
		if (input_skb) {
			// debug printk("Freeing #%lu %p\n", ticket_good_drops, input_skb);
			kfree_skb(input_skb);
			input_skb=NULL;
			ticket_good_drops++;
			drop_idx=0;
		}

		// handle delay characteristics.
		if (mu_s_diff(&tv_start, &tv_now)>= (1000*routing_states[cur_delay].duration)) {
			cur_delay = change_delay(cur_delay, &tv_start, &tv_now, &drop_idx);
		}
	}

	/* cleanup the thread, quit. */
	lock_kernel(); /* get BKL; will muck with junk */
	packet_thread.thread = NULL; /* Thread is to-be gone */
	mb(); /* propogate flags */
	up(&packet_thread.control_sem); /* let cleanup_module know. */
	/* In theory; the kernel_thread that called clone() now calls do_exit() here,
	 * which releases BKL for me. */
	return;
}

/* -----------------------------------------------------------------------------
 * start up module. This function:
 *     steals netif_rx,
 *     brings up all interfaces to use,
 *     starts dump thread.
 * -------------------------------------------------------------------------- */
int ticket_route_init(void) {
	printk(KERN_INFO "Loading TICKET route module...\n");
	old_netif_rx = netif_rx; /* set first so failures can just call ticket_exit */
#if LINUX_VERSION_CODE>=0x020420
	old_netif_receive = netif_receive_skb_orig;
#endif

	/* bring up all interfaces */
	if (!ticket_ifupall()) {
		ticket_route_exit();
		/* failure */
		return 1;
	}

	/* create new kernel thread. Zero everything first */
	memset(&packet_thread, '\0', sizeof(struct ew_thread));
	/* set up packet queue too */
	skb_queue_head_init(&ticket_skbs);

	/* Initialize everything as locked; next down will block. */
	init_MUTEX_LOCKED(&packet_thread.control_sem); /* control creation/destruction of thread */
	init_MUTEX_LOCKED(&packet_thread.packet_wait_sem); /* thread sleeps here waiting for packets */

	/* create the new thread by running a task through keventd */
	/* initialize the task queue structure */
	INIT_LIST_HEAD(&packet_thread.taskq.list);
	packet_thread.function = &ticket_route_thread;
	packet_thread.taskq.routine = ew_thread_create;
	packet_thread.taskq.data = &packet_thread;

	/* schedule its execution, then let it run 'till init is complete... */
	schedule_task(&packet_thread.taskq);
	down(&packet_thread.control_sem);

	/* success! */
	return 0;
}


/* -----------------------------------------------------------------------------
 * clean up module, exit. This function:
 *     resets netif_rx,
 *     stops dump thread,
 *     brings down all used interfaces.
 * -------------------------------------------------------------------------- */
void ticket_route_exit(void) {
	int idx;
	struct sk_buff *skb;

	printk(KERN_INFO "Unloading TICKET module... resetting netif_rx(s).\n");
	netif_rx = old_netif_rx;
#if LINUX_VERSION_CODE>=0x020420
	netif_receive_skb = old_netif_receive;
#endif

	/* --------------------------------------------------------------------------------
	 * kill the kernel thread. must be _AFTER_ reset of netif_rx;
	 * Otherwise we could get more packets which would be messy.
	 * -------------------------------------------------------------------------------- */
	if (packet_thread.thread != NULL) {
		lock_kernel(); /* get BKL, will muck with process data. */
		init_MUTEX_LOCKED(&packet_thread.control_sem); /* We'll wait for thread to stop cleanly.. */
		packet_thread.die = 1; /* tell thread to die... */
		mb(); /* Make sure flags visible on all CPUs. */
		up(&packet_thread.packet_wait_sem); /* Wake thread from packet wait, let it die cleanly */
		down(&packet_thread.control_sem); /* Block till thread function completes cleanup */
		unlock_kernel(); /* release the BKL */
		kill_proc(2, SIGCHLD, 1); /* Now thread is a zombie; Make keventd clean it up. */
	}

	/* --------------------------------------------------------------------------------
	 * now close all devices we opened, require netlink lock
	 * -------------------------------------------------------------------------------- */
	rtnl_lock();
	for (idx=0; idx<ETHERNET_IF_COUNT && ew_eth[idx]!=NULL; idx++) {
		// Only close if that's how it was before. Otherwise, it was open. Don't touch.
		if (ew_eth_oldstatus[idx]==NET_DEVICE_CLOSED) {
			printk(KERN_INFO "Closing %s.\n", ew_eth[idx]->name);
			dev_close(ew_eth[idx]);
		}
		dev_put(ew_eth[idx]);
		ew_eth[idx] = NULL;
	}
	rtnl_unlock();

	// --------------------------------------------------------------------------------
	// clean up rest of memory
	//
	// Remember to free anything in a list (allocates memory per item)
	//    (we don't have any ATM, but may later)
	// --------------------------------------------------------------------------------
	// free mapping, if we have one allocated
	if (map_addrs!=NULL) {
		free_mapping(map_addrs);
		map_addrs=NULL;
	}

	// Free any packets we haven't gotten around to processing.
	if (skb_queue_len(&ticket_skbs)>0) {
		printk("Throwing away %u unprocessed packets; didn't get to them.\n", skb_queue_len(&ticket_skbs));
		printk("Correctly dropped %lu evil packets (non-forwardable).\n", ticket_good_drops);
		printk("Unfortuantely dropped %lu holy packets (backlog handling).\n", ticket_evil_drops);
		do {
			if ((skb = skb_dequeue(&ticket_skbs))!=NULL) {
				kfree_skb(skb);
			}
		} while (skb_queue_len(&ticket_skbs)>0);
	}
}

/* -----------------------------------------------------------------------------
 * Other stuff; see linux/module.h for more information, or `man modinfo`
 * -------------------------------------------------------------------------- */
#ifdef MODULE
module_init(ticket_route_init);
module_exit(ticket_route_exit);

MODULE_AUTHOR("Eric Weigle <ehw@lanl.gov>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TICKET route kernel code");
MODULE_PARM(pass_through, "i");
MODULE_PARM_DESC(pass_through, "(boolean) Pass packets through to original netif_rx when done?");
#endif

