/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// User networking code;
//
// Wrappers for UDP socket initialization, recvfrom and sendto.
// #############################################################################

#include "network.h"
#include "net_headers.h"
#include "misc.h"
#include "config.h"

// =============================================================================
// FILE-LOCAL function: handles error conditions for this file.
// =============================================================================
void local_perror(char *fail_msg, char *loc_msg, int loc_perror) {
	if (fail_msg) {
		fprintf(stderr, "(context '%s') ", fail_msg);
	}

	if (loc_msg) {
		fprintf(stderr, "Observed error '%s'", loc_msg);
	} else {
		fprintf(stderr, "Observed miscellaneous error");
	}

	if (loc_perror) {
		fprintf(stderr, ": '%s'\n", strerror(errno));
	} else {
		fprintf(stderr, "\n");
	}
}


// =============================================================================
// Set up UDP socket. Return the filehandle.
// Pass in REAL server_name to create a client (clients SEND)
//         NULL server_name for a server (servers RECEIVE)
// If want to attempt to set up buffer size,
//    pass in nonzero 'bufsize'
// =============================================================================
int create_udp_socket(char *server_name, struct sockaddr_in *serv_addr, int bufsize, int verbosity) {
	struct sockaddr_in cli_addr;
	struct hostent *host_ent;
	int connfd;

	// error check.
	if (server_name==NULL && serv_addr==NULL) {
		fprintf(stderr, "Bad(NULL) parameters passed to create a UDP socket!\n");
		return 0;
	}

	// both client and server have to set up server address
	bzero((char *) serv_addr, sizeof(*serv_addr)); // always initialize to zeroes
	serv_addr->sin_family = AF_INET;
	serv_addr->sin_port = htons(SERVER_TICKET_PORT); // this is different for each program!

	if (server_name) {
		// client sends to server's address, which must first be looked up.
		if ((host_ent = gethostbyname(server_name)) == NULL) {
			perror("Client(gethostbyname)");
			return 0;
		}
		memcpy(&(serv_addr->sin_addr.s_addr), host_ent->h_addr, host_ent->h_length);
	} else {
		// server accepts traffic at any local address.
		serv_addr->sin_addr.s_addr = htonl(INADDR_ANY);
	}

	// Opens a UDP socket (an internet datagram socket)
	if ((connfd=socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("Client(cannot open socket)");
		return 0;
	}

	if (server_name) {
		// client binds any local address
		bzero((char *) &cli_addr, sizeof(cli_addr)); //  zero out
		cli_addr.sin_family = AF_INET;
		cli_addr.sin_addr.s_addr = htonl(INADDR_ANY); //  whatever
		cli_addr.sin_port = htons(0);
		if (bind(connfd, (struct sockaddr *) &cli_addr, sizeof(cli_addr)) < 0) {
			perror("Client(cannot bind local address)");
			return 0;
		}

		// clients SEND buffer
		if (bufsize>0 && setsockopt(connfd, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof(int))!=0) {
			fprintf(stderr, "Soft-failing; couldn't set send buffer size %d, because '%s'\n", bufsize, strerror(errno));
			fprintf(stderr, "Try `echo %d > /proc/sys/net/core/wmem_max`\n", bufsize);
		}
	} else {
		// server binds the specific address given above.
		if (bind(connfd, (struct sockaddr *) serv_addr, sizeof(*serv_addr)) < 0) {
			perror("Server(bind local address)");
			return 0;
		}

		// servers RECEIVE buffer
		if (bufsize>0 && setsockopt(connfd, SOL_SOCKET, SO_RCVBUF, &bufsize, sizeof(int))!=0) {
			fprintf(stderr, "Soft-failing; couldn't set receive buffer size %d, because '%s'\n", bufsize, strerror(errno));
			fprintf(stderr, "Try `echo %d > /proc/sys/net/core/rmem_max`\n", bufsize);
		}
	}

	if (verbosity > 0) {
		printf("I've set up a UDP socket for server \"%s\"\n", server_name);
	}

	return connfd;
}


// =============================================================================
// FILE-LOCAL function: Called to handle timeouts in receive via select.
// Failure cases will probably lead to the death of the caller.
// =============================================================================
int recvtimeo(int fd) {
	fd_set rset;
	struct timeval tv;

	// zero the file descriptor set, and then add fd to it.
	FD_ZERO(&rset);
	FD_SET(fd, &rset);

	// set up timeout
	tv.tv_sec=TICKET_RECV_TIMEOUT;
	tv.tv_usec=0;

	// Call select with a set of file descriptors to watch for readability
	// changes, no file descriptors to watch for writability or exceptions
	return (select(fd+1, &rset, NULL, NULL, &tv)<=0);
}


// =============================================================================
// A simple call to recvfrom encapsulated because we do it all over the place.
// Slightly simplifies code in callers.
// =============================================================================
int full_recvfrom(int fd, char *buf, int length, int flags, struct sockaddr_in *addr, char *fail_msg) {
	int addrlen;
	int newlength;

	// if we don't want to block, then check for read status with select.
	if (flags!=FLAG_BLOCK && recvtimeo(fd)) {
		local_perror(fail_msg, "Timed out whilst waiting in receive", 0);
		return 0;
	}

	// Ok, now we should have data to read...
	addrlen = sizeof(*addr);
	// NOTE: currently using a '<' check here; allows us to accept variable sized packets
	// may want to use a '!=' and pass in an EXPECTED length, for more strict checking.
	if ((newlength = recvfrom(fd, buf, length, 0, (struct sockaddr *)addr, (socklen_t *)&addrlen)) < (int)sizeof(int)) {
		local_perror(fail_msg, "Recvfrom error", 1);
		return 0;
	}

	return length;
}


// =============================================================================
// A simple call to sendto encapsulated because we do it all over the place.
// Slightly simplifies code in callers. I don't deal with timeouts here,
// this *should* succeed every time... if not you've got other problems!
// =============================================================================
int full_sendto(int fd, char *buf, int length, struct sockaddr_in *addr, char *fail_msg) {
	if (sendto(fd, buf, length, 0, (struct sockaddr *)addr, sizeof(*addr)) != length) {
		local_perror(fail_msg, "Sendto error", 1);
		return 0;
	}

	return 1;
}

