/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// ``Forger'' program: Send fake dump packets over the network.
//
// Used for testing: Fire up a couple of these sending to one machine to see what
// kind of data rate they can handle; if it's not fast enough you know how many
// machines between which you need to stripe your data.
// #############################################################################
#include <arpa/inet.h>   // inet_pton, inet_ntoa
#include <time.h>        // nanosleep, etc
#include <stdlib.h>      // malloc, free
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
#include "../common/config.h"
#include "../common/dump.h"
#include "../common/network.h"
#include "../common/doubly.h"

// =============================================================================
// lists holding information on how many packets to send,
// and how long to wait between sending each packet.
//
// Relies on the fact that void * and int are the same size.
// =============================================================================
struct dlinked_list count_list;
struct dlinked_list sleep_list;

char *hostname=NULL;
int verbosity;
int bufsize;


// =============================================================================
// FUNCTION DECLARATIONS
// =============================================================================
char *strn_alloc_cpy(char **dst, const char *src, size_t max_size);
int  list_add_number(struct dlinked_list *cur_list, char *number);
void cleanup_memory(void);
void do_forging(struct sockaddr_in *serv_addr);
void parse_command_line(int argc, char *argv[]);
void usage(char *error);


// =============================================================================
// Destroy allocated memory.
// =============================================================================
void cleanup_memory(void) {
	int idx;

	if (hostname) {
		free(hostname);
		hostname=NULL;
	}

	// only two lists to kill.
	while (count_list.count>0) {
		(void)dlist_pop_head(&count_list, (void **)&idx);
	}
	while (sleep_list.count>0) {
		(void)dlist_pop_head(&sleep_list, (void **)&idx);
	}
}


// =============================================================================
// Usage: how to call this program.
// =============================================================================
void usage(char *error) {
	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: ./forger <-v -v+> <-b [bufsize]> [-h hostname] [-p packets] [-s sleeptime] <-p packets> <-s sleeptime> ...\n\n");

	printf("    -v, -v -v, -v -v -v: run verbosely, very verbosly, and in debugging mode\n");
	printf("    -b=[bufsize]: set the socket buffer size to this value.\n"); 
	printf("    hostname: To which host we send packets.\n");
	printf("    packets: number of packets to forge\n");
	printf("    sleeptime: how long to pause (ms) between packet transmissions.\n");
	printf("Any number of pairs of packets and sleeptimes may be used.\n");
	printf("The program terminates after sending all packets.\n\n");
	printf("For example: `./forger -vv 1000000000 0`:\n");
	printf("             sends a billion packets as fast as possible.\n");

	cleanup_memory();

	exit(1);
}


// =============================================================================
// Allocate space and then copy the string from src to dst.
// TODO: move to common code... (?)
// =============================================================================
char *strn_alloc_cpy(char **dst, const char *src, size_t max_size) {
	int len;

	// error check.
	if (dst==NULL || src==NULL) {
		printf("NULL string passed to strn_alloc_copy.\n");
		return NULL;
	}

	// length of string + terminating null
	len = strlen(src)+1;
	if (len>max_size) {
		printf("Long string passed to strn_alloc_copy.\n");
		return NULL;
	}

	// otherwise, allocate and copy.
	if ((*dst = calloc(1, len))==NULL) {
		printf("Failed to allocate %d bytes for string copy\n", len);
		return NULL;
	}
	strncpy(*dst, src, len);

	return(*dst);
}


// =============================================================================
// Parse `number' and add to the list.
// =============================================================================
int list_add_number(struct dlinked_list *cur_list, char *number) {
	int idx;

	// verify it's a number.
	for (idx=0; idx<strlen(number); idx++) {
		if (!isdigit(number[idx])) {
			printf("Argument '%s' is not numeric.\n", number);
			usage("Input error");
		}
	}

	// then actually get value, shove on list.
	idx=0;
	if (sscanf(number, "%d", &idx)!=1 || idx<=0) {
		usage("Oops, negative or zero parameter... count and duration must be postive");
	}
	dlist_push_tail(cur_list, (void *)idx);

	return (idx);
}


// =============================================================================
// Parse the command line.
// =============================================================================
void parse_command_line(int argc, char *argv[]) {
	int val, loptidx=0;
	struct option long_options[] = {
		/* name, has_arg, int*flag, int val) */
		{"verbose",  0, 0, 'v' },
		{"bufsize",  1, 0, 'b' },
		{"hostname", 1, 0, 'h' },
		{"packets",  1, 0, 'p' },
		{"sleep",    1, 0, 's' },
		{0,          0, 0,  0  }
	};

	// initialize.
	verbosity=0;
	bufsize=0;

	// then get options.
	for ( ; (val=getopt_long(argc, argv, "vb:h:p:s:", long_options, &loptidx))!=-1; ) {
		switch (val) {
			case 'v':
				verbosity++;
				break;

			case 'b':
				if (sscanf(optarg, "%d", &bufsize)!=1) {
					usage("Failed scan of bufsize string!");
				}
				if (bufsize<=0) {
					usage("Scanned bad (negative or zero) bufsize string!");
				}

				break;

			case 'h':
				// printf("option h with value `%s'\n", optarg);
				if (strn_alloc_cpy(&hostname, optarg, 512)==NULL) {
					usage("Failed to copy host name; try another (?)\n");
				}
				break;

			case 'p':
				// printf("option p with value `%s'\n", optarg);
				(void)list_add_number(&count_list, optarg);
				break;

			case 's':
				// printf("option s with value `%s'\n", optarg);
				(void)list_add_number(&sleep_list, optarg);
				break;

			case '?':
				usage("Unrecognized option...");
				break;

			default:
				usage("getopt returned unhandled character code.");
		}
	}

	if (optind < argc) {
		usage("Excess options on command line!");
	}

	exit(0);

	// error check
	if (sleep_list.count != count_list.count) {
		usage("Need *pairs* of numbers, packet count and sleep time");
	}

	if (verbosity>0) {
		int countval, sleepval, idx;

		printf("Sending to host '%s', using the following table:\n", hostname);
		printf("\tPacket Count   Sleep Time\n");
		for (idx=0; idx<count_list.count; idx++) {
			(void)dlist_pop_head(&count_list, (void **)&countval);
			(void)dlist_pop_head(&sleep_list, (void **)&sleepval);
			printf("\t%12d %12d\n", countval, sleepval);
			(void)dlist_push_tail(&count_list, (void *)countval);
			(void)dlist_push_tail(&sleep_list, (void *)sleepval);
		}
	}

	exit(1);
}


// =============================================================================
// Do the actual forging  of packets.
// =============================================================================
void do_forging(struct sockaddr_in *serv_addr) {
	int sockfd=0;
	dump_packet packet;
	struct timespec sleeptime;
	unsigned long long seqno=0;
	int sleep_every, ret_val;
	int cur_count, cur_sleep;

	bzero(&packet, sizeof(dump_packet));
	
	// now run the tables passed in.
	while (count_list.count>0) {
		(void)dlist_pop_head(&count_list, (void **)&cur_count);
		(void)dlist_pop_head(&sleep_list, (void **)&cur_sleep);

		if (verbosity>0) {
			printf("Moving to next count/duration (%d/%d)\n", cur_count, cur_sleep);
		}

		// We use 'nanosleep' to handle pauses between packets. Due to linux context
		// switch times, etc. this has maximal accuracy of about 10ms, a problem.
		// This logic makes us sleep every 'N' times for 'N*sleep_time' when 'sleep_time'
		//   is less than 10ms; that is, since we can't sleep less than 10ms, when we
		//   want that effect we sleep for an appropriate period every so often.
		// Example: want to sleep 4ms but can't. So sleep for 12 ms, once every 3 loops.
		for (sleep_every=1; (sleep_every*cur_sleep)<10 && cur_sleep>0; sleep_every++) {
			;
		}
		cur_sleep *=sleep_every;

		// send 'count' packets, pausing for 'duration'
		for ( ; cur_count>0; cur_count--) {
			packet.seqno = seqno++;
			// send it off...
			if ((ret_val = sendto(sockfd, &packet, sizeof(dump_packet), 0, (struct sockaddr *)serv_addr, sizeof(*serv_addr)))!=sizeof(dump_packet)) {
				printf("failed sendto: ret_val is %d, because '%s'\n", ret_val, strerror(errno));
			}

			// Then sleep:
			if (cur_count%sleep_every==0 && cur_sleep>0) {
				sleeptime.tv_sec = 0;
				sleeptime.tv_nsec = cur_sleep*1000000; // 1 millisecond in nanoseconds
				nanosleep(&sleeptime, NULL);
			}
		}
	}
}


// =============================================================================
// The main routine. Basically, this just repeatedly sends an empty
// dump_packet we've created, with the sequence number incremented each time.
// =============================================================================
int main(int argc, char *argv[]) {
	struct sockaddr_in serv_addr;

	// initialize, start up
	memset(&count_list, sizeof(struct dlinked_list), '\0');
	memset(&sleep_list, sizeof(struct dlinked_list), '\0');
	parse_command_line(argc, argv);

	if (!create_udp_socket(hostname, &serv_addr, bufsize, verbosity)) {
		printf("Couldn't create socket\n");
		return(1);
	}

	do_forging(&serv_addr);

	// done! clean up and quit.
	cleanup_memory();

	return 0;
}

