/*
###############################################################################
# This file is distributed as part of TICKET, but was written independently.
# It is Copyright (C) 2001 by Eric H. Weigle.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at weigle@cs.wisc.edu,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following website for updated information or other means by which he can
# be contacted:
# 	http://www.losalamos.com/ehw/
#
###############################################################################
*/
// #############################################################################
// Basic linked hash table code, header file.
// #############################################################################

#ifndef EW_HASHTABLE_H
#define EW_HASHTABLE_H

// =============================================================================
// Hash structures.
// =============================================================================
struct hashtable_item {
	void *item;
	unsigned int magic;
	struct hashtable_item *next;
};

struct hashtable {
	int nitems; // how many items are in the table
	int size; // size of the table.

	unsigned int (*hashfn)(void *item, int size);
	int (*compfn)(const void *item_a, const void *item_b);
	int (*printfn)(FILE *file, const void *item);

	struct hashtable_item **table; // the table itself.
};

// =============================================================================
// Functions.
// =============================================================================

// call to create a hastable of size 'size' (allocate memory)
// pass in an already created struct hashtable.
// returns -1 on stupid user or failure, 1 on success
int hash_create(struct hashtable *table, int size, unsigned int (*hashfn)(void *, int), int (*compfn)(const void *, const void *), int (*printfn)(FILE *file, const void *));

// destroy the hash table (free memory) Have to destroy structure yourself.
// returns -1 on stupid user/failure, 1 on success.
int hash_destroy(struct hashtable *table);

// remove the given item from the table, if it exists.
// If offset and magic are not NULL, sets them to be the offset in the list and the magic in the hash item (number of collisions)
// returns: -1 on failure, index on sucess
int hash_remove(struct hashtable *table, void *item, int *offset, int *magic);

// returns the item index from the table, if it exists, WITHOUT REMOVING IT.
// If offset and magic are not NULL, sets them to be the offset in the list and the magic in the hash item (number of collisions)
// returns: -1 on failure, index on success.
int hash_find(struct hashtable *table, void *item, int *offset, int *magic);

// add the given item to the table, if it exists.
// returns: magic number of item.
unsigned int hash_insert(struct hashtable *table, void *item, unsigned int magic);

// clear the hashtable, return an array of all the hashed items and all their magic.
// Allocates 'items' and 'magics'; user must free!
// returns: -1 on failure, count on success.
int hash_unhash(struct hashtable *table, void ***items, int **magics);

// print the table. pass in a callback to print a single element.
// returns: -1 on failure, 1 on success
int print_table(FILE *file, struct hashtable *table);

// =============================================================================
// Varius useful hashes.
// =============================================================================

// Do a web search for "g_str_hash X31_HASH" if you want to know more.
unsigned int string_hash(void *item, int size);

// Xor first and second 16 bits, divide by 65537 (a prime, just over 65536)
unsigned int ipaddr_hash(void *item, int size);
int ipaddr_cmp(const void *item_a, const void *item_b);
int ipaddr_print(FILE *file, const void *item);

unsigned int macaddr_hash(void *item, int size);
int macaddr_cmp(const void *item_a, const void *item_b);
int macaddr_print(FILE *file, const void *item);

// =============================================================================
// I don't hash IP protocols because there are only 256 of 'em, nor do I hash
// TCP/UDP ports because there are only 64K of 'em. For those, just allocate a
// 256B buffer and a 64KB buffer and hash via modulo.
// =============================================================================

#endif

