/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// Reader code; takes output saved to disk by writer, taken from kernel dump.
// #############################################################################
#include <netinet/in.h>       // ntohs, ntohl
#include <errno.h>            // used when trying to mkdir
#include <fcntl.h>            // open()
#include <stdlib.h>           // malloc(), close()
#include <string.h>           // strncpy, strncmp
#include <strings.h>          // bzero() and others
#include <unistd.h>           // close()
#include <stdio.h>
#include "read_common.h"
#include "../common/doubly.h"


// =============================================================================
// Usage: how to call this program.
// =============================================================================
void reader_usage(char *error) {
	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: ./reader [-c=cpu_speed]  [-s=trace_start] file1 [file2 ...]\n");
	printf("    * optional CPU speed is in hertz; try this:\n");
	printf("      `cat /proc/cpuinfo|grep 'cpu MHz'|sed -e 's/.* \\(.*\\)/\\1 100000*p/'|dc`\n");
	printf("    * optional trace start time is in seconds since the UNIX epoch;\n");
	printf("        you can get this by looking at the close time of a prior file,\n");
	printf("        or the system_info.txt file created by the writer program.\n");

	exit(1);
}


// =============================================================================
// Parse the command line.
// =============================================================================
void parse_reader_command_line(int argc, char *argv[], int *file_count, char ***filenames, unsigned long *cpu_speed, unsigned long *trace_start) {
	int idx, length;

	if (argc<2) {
		reader_usage("Too few arguments.");
	}

	// then get files list.
	*file_count = argc-1;

	if ((*file_count)<=0) {
		reader_usage("Failed to pass in file name[s] on command line.\n");
	}

	// get copy of all filenames.
	if ((*filenames = (char **)malloc((*file_count)*sizeof(char *)))==NULL) {
		perror("Failed to allocate file name pointers!\n");
		exit(1);
	}
	for ((*file_count)=0,idx=1; idx<argc; idx++) {
		if (argv[idx][0] != '-') {
			// Not an option. Add filename to list.
			length = strlen(argv[idx]) + 1;
			if (((*filenames)[*file_count] = (char *)malloc((length)*sizeof(char)))==NULL) {
				perror("Failed to allocate file name pointers!\n");
				exit(1);
			}
			memcpy((*filenames)[*file_count], argv[idx], length);
			(*file_count)++;
		} else {
			// TODO???? Verbosity handling here.
			// an option. parse it out.
			if (strlen(argv[idx])<4) {
				reader_usage("Short option!\n");
			}
			if ((length = atol(&(argv[idx][3])))<1000000) {
				reader_usage("Bad option (<1 million)");
			}

			// save it to parameter.
			switch(argv[idx][1]) {
				case 'c': case 'C': if (cpu_speed  !=NULL) { *cpu_speed   = length; } break;
				case 's': case 'S': if (trace_start!=NULL) { *trace_start = length; } break;
				default: reader_usage("Bad option on command line!");
			}
		}
	}
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void ntoh_collected_header(struct collected_headers_ticket *hdr) {
	// hdr->length= ntohs(hdr->length); // OOPs! length is already little-endian.
	hdr->toslen   = ntohs(hdr->toslen);
	hdr->ip_src   = ntohl(hdr->ip_src);
	hdr->ip_dst   = ntohl(hdr->ip_dst);
	hdr->ip_len   = ntohs(hdr->ip_len);
	hdr->src_port = ntohs(hdr->src_port);
	hdr->dst_port = ntohs(hdr->dst_port);

	// This weird conversion is because the header length
	// is really in 32-bit words. The collection code copied it
	// from the last 4 bits of the packet field into the first char of
	// the local field. Convert to bytes by multiplying it by 4.
	// 20020607: removed, already done by kernel code.
	// hdr->ip_hdr_len = ((hdr->ip_hdr_len)&0xF)<<2;

	// the "upper_len" field is different for TCP and UDP
	if (hdr->protocol==UDP_HDR_PROTOCOL) {
		// packet length in NBO. convert to HBO.
		hdr->upper_len.udp_len = ntohs(hdr->upper_len.udp_len);
	} else if (hdr->protocol==TCP_HDR_PROTOCOL) {
		// header length, in 32-bit words. Convert to bytes via shift left 2 or multiply by 4
		hdr->upper_len.udp_len = ((unsigned short)hdr->upper_len.tcp_len.len) << 2;
		// NOTE: I've seen some odd things with this next line of code on some packets;
		// Some incongruity between header lengths screws up here. I haven't tracked it down
		// quite yet... Assuming it's a buggy TCP/IP stack of some sort, but not sure.
		// Then, we really want the actual packet length, not the length of the header (like UDP)
		// In short, I saw negative or too-small numbers.
		//     hdr->upper_len.udp_len = hdr->ip_len - hdr->upper_len.udp_len;
	}
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void ntoh_processed_header(struct processed_headers *hdr) {
	hdr->timestamp.tv_sec  = ntohl(hdr->timestamp.tv_sec);
	hdr->timestamp.tv_nsec = ntohl(hdr->timestamp.tv_nsec);
	hdr->length            = ntohs(hdr->length);

	hdr->mac_src = ntohl(hdr->mac_src);
	hdr->mac_dst = ntohl(hdr->mac_dst);
	hdr->toslen  = ntohs(hdr->toslen);

	hdr->ip_src = ntohl(hdr->ip_src);
	hdr->ip_dst = ntohl(hdr->ip_dst);
	hdr->ip_len = ntohs(hdr->ip_len);

	hdr->src_port  = ntohs(hdr->src_port);
	hdr->dst_port  = ntohs(hdr->dst_port);
	hdr->upper_len = ntohs(hdr->upper_len);
}


// =============================================================================
// Perform endian conversions on all fields in our collected headers.
// =============================================================================
void hton_processed_header(struct processed_headers *hdr) {
	hdr->timestamp.tv_sec  = htonl(hdr->timestamp.tv_sec);
	hdr->timestamp.tv_nsec = htonl(hdr->timestamp.tv_nsec);
	hdr->length            = htons(hdr->length);

	hdr->mac_src = htonl(hdr->mac_src);
	hdr->mac_dst = htonl(hdr->mac_dst);
	hdr->toslen  = htons(hdr->toslen);

	hdr->ip_src = htonl(hdr->ip_src);
	hdr->ip_dst = htonl(hdr->ip_dst);
	hdr->ip_len = htons(hdr->ip_len);

	hdr->src_port  = htons(hdr->src_port);
	hdr->dst_port  = htons(hdr->dst_port);
	hdr->upper_len = htons(hdr->upper_len);
}


// // =============================================================================
// // Compare packet sequence numbers; for sorting.
// // =============================================================================
// static int pkt_seqno_compare(void *pkt_a, void *pkt_b) {
// 	return (((struct dump_packet *)pkt_b)->seqno - ((struct dump_packet *)pkt_a)->seqno);
// }


// =============================================================================
// read processed data from the file. For driving analyzer code.
// =============================================================================
void do_generic_read(char **filenames, int file_count, size_t struct_size, int (*analysis_fn)(void *data, char *cur_file)) {
	void *data;
	int idx, infd;
	unsigned long long nread=0;

	if ((data = calloc(1, struct_size))==NULL) {
		printf("Failed allocation of %d bytes in do_generic_read()\n", struct_size);
		exit(1);
	}

	// read all files.
	for (idx=0; idx<file_count; idx++) {
		// open file.
		if ((infd = open(filenames[idx], O_RDONLY))<0) {
			printf("FAILURE! Can't open input file '%s' because '%s'\n", filenames[idx], strerror(errno));
			exit(1);
		}

		// Ready to read
		for ( ; read(infd, data, struct_size)==struct_size; nread++) {
			// ----------------------------------------
			// put your analysis code here.
			// ----------------------------------------
			analysis_fn(data, filenames[idx]);
		}
		// close files
		close(infd);

		// give stats.
		printf("After having read file '%s', nread is %lld\n", filenames[idx], nread);
	}
}

