/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// A simple program to read in the first and last records from a file, and to
// print out their sequence number and timestamp.
//
// Useful to find out how much data in each file, and what time range each
// file holds.
// #############################################################################
#include <stdio.h>
#include <fcntl.h>            // open()
#include <errno.h>            // used when trying to mkdir
#include <string.h>           // strncpy, strncmp
#include <strings.h>          // bzero() and others
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include "read_common.h"
#include "print.h"

// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	int infd, idx, file_count=0;
	char **filenames=NULL;
	unsigned long cpu_speed=0;
	struct collected_headers_ticket first_pkt, last_pkt;
	struct stat stat_buf;
	off_t offset;

	// parse command line.
	parse_reader_command_line(argc, argv, &file_count, &filenames, &cpu_speed, NULL);

	// read all files.
	for (idx=0; idx<file_count; idx++) {
		printf("Reading input file '%s'\n", filenames[idx]);

		// open file.
		if ((infd = open(filenames[idx], O_RDONLY))<0) {
			printf("FAILURE! Can't open input file '%s' because '%s'\n", filenames[idx], strerror(errno));
			exit(1);
		}

		if (fstat(infd, &stat_buf)<0) {
			printf("Failed to stat '%s' because '%s'\n", filenames[idx], strerror(errno));
			close(infd);
			exit(1);
		}
		if (stat_buf.st_size < 2*sizeof(struct collected_headers_ticket)) {
			printf("Cannot read file; size too small. (%lu < %lu)\n", stat_buf.st_size, (unsigned long)sizeof(struct collected_headers_ticket));
			close(infd);
			exit(1);
		}

		// Read first packet
		if (read(infd, &(first_pkt), sizeof(struct collected_headers_ticket))==sizeof(struct collected_headers_ticket)<=0) {
			printf("Couln't read first packet!\n");
			exit(1);
		}

		// This seek doesn't work correctly on abruptly terminated files.
		if ((stat_buf.st_size % sizeof(struct collected_headers_ticket))!=0) {
			printf("Warning: file size (%lu) is not multiple of size of a collected packet (%lu)!\n", stat_buf.st_size, (unsigned long)sizeof(struct collected_headers_ticket));
		}

		offset = -1 * (sizeof(struct collected_headers_ticket)+ (stat_buf.st_size%sizeof(struct collected_headers_ticket)));
		if (lseek(infd, offset, SEEK_END)<offset) {
			printf("Couln't seek to end!\n");
			exit(1);
		}

		// read last packet
		if (read(infd, &(last_pkt), sizeof(struct collected_headers_ticket))==sizeof(struct collected_headers_ticket)<=0) {
			printf("Couln't read last packet!\n");
			exit(1);
		}

		// print first and last packets.
		if (cpu_speed>0) {
			printf("%f\n", ((double)first_pkt.timestamp/(double)cpu_speed));
			printf("%f\n", ((double)last_pkt.timestamp/(double)cpu_speed));
		} else {
			printf("%lld\n", first_pkt.timestamp);
			printf("%lld\n", last_pkt.timestamp);
		}

		// close files
		close(infd);
	}

	// clean up allocations
	for (idx=0; idx<file_count; idx++) {
		free(filenames[idx]);
	}
	free(filenames);

	exit(0);
}

