#!/usr/bin/perl -w
use strict;
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#

###############################################################################

# #############################################################################
# This script installs the TICKET kernel code into a properly configured Linux
# kernel source tree, the location of which is a command-line parameter we
# store in $source_tree (for example, $source_tree=/usr/src/linux-2.4.17).
#
# The idea is to:
#     1) Determine the location of things;
#         Linux kernel source directories ('kernel' and 'common')
#         TICKET kernel source.
#     2) Copy them;
#         kernel: '$ticket_tree/* to $source_tree/ticket/
#         common: '$common_tree/* to $source_tree/ticket/common/
#
# Then patch up other kernel files to connect with the TICKET code.
# In particular we: 
#     add a call to our 'main' function in $source_tree/boot/init/main.c
#     modify the version string in $source_tree/boot/init/version.c
#     set up changes to netif_rx so packets are sent directly to TICKET.
#
# #############################################################################

my ($source_tree, $ticket_tree, $common_tree) = ("","","");
my $mode='i'; # install: i, uninstall: u
my $verbosity=0;
my $TICKET_VERSION="alpha release 0.1";
my @lines;


# =============================================================================
# Print all messages to stdout and to an install status file too.
# =============================================================================
sub log_msg() {
	my ($msg_priority, $msg) = @_;

	if ($verbosity>=$msg_priority) {
		print $msg;
	}

	print LOGFILE "$msg_priority\t$msg";
}


# =============================================================================
# Tell user how to call program.
# =============================================================================
sub usage() {
	my ($error) = @_;

	if ($error ne "") {
		&log_msg(0, "ERROR!:\n    $error\n\n");
	}

	&log_msg(0, "Usage: $0 <-vvv...> <-i|-u> [-k=linux_kernel_source_dir] <-t=ticket_kernel_source_dir>\n\n");

	&log_msg(0, "-v,-vv,-vvv: Be verbose, very verbose, etc.\n");
	&log_msg(0, "-i|-u: install or uninstall; default to '-i' for installation.\n");
	&log_msg(0, "While you must specifiy the linux kernel source directory into which we\n");
	&log_msg(0, "    install or uninstall, the ticket directory is usually not required\n");

	&log_msg(0, "    (this program should be able to correctly derive its location).\n\n");
	&log_msg(0, "If you are following the user guide, you probably just need to call\n");
	&log_msg(0, "    $0 -k=/usr/src/linux-ticket/\n");

	exit 1;
}


# =============================================================================
# Parse command line
# 20021011- Allow parameters in arbitrary order; flag-ized everything.
# =============================================================================
sub parse_cmd_line() {
	my $idx;

	for ($idx=0; $idx<=$#ARGV; $idx++) {
		if ($ARGV[$idx] =~ /-[vV]+/) {
			# verbosity
			$verbosity = length($ARGV[$idx])-1;
		} elsif ($ARGV[$idx] =~ /-[uU]/) {
			# mode uninstall
			$mode = 'u';
		} elsif ($ARGV[$idx] =~ /-[iI]/) {
			# mode install
			$mode = 'i';
		} elsif ($ARGV[$idx] =~ /-[kK]=./) {
			($source_tree) = ($ARGV[$idx] =~ /...(.*)/);
		} elsif ($ARGV[$idx] =~ /-[tT]=./) {
			($ticket_tree) = ($ARGV[$idx] =~ /...(.*)/);
		}

	}

	# then a bunch of error checks.
	if ($mode ne 'u' && $mode ne 'i') {
		&usage("Bad mode ($mode)!");
	}

	# ONLY Installation can have an optional ticket directory.
	if ($mode eq 'u' && $ticket_tree ne "") {
		&log_msg(0, "Ignoring unused TICKET directory $ticket_tree..\n");
		$ticket_tree = "";
	}

	# have to have a kernel source directory...
	if ($source_tree eq "") {
		&usage("Need a source directory.");
	}

	# and it must be valid.
	if (!(-e $source_tree) || !(-d $source_tree)) {
		&usage("'$source_tree' does not appear to both exist and be a regular directory.");
	}

	# if there's a TICKET tree, it must also be valid.
	if ($ticket_tree ne "") {
		if (!(-e $ticket_tree) || !(-d $ticket_tree)) {
			&usage("'$ticket_tree' does not appear to both exist and be a regular directory.");
		}
	}
}


# =============================================================================
# Given a path, see if it's a link and if so refer to what it is pointing.
# =============================================================================
sub make_linkless_path() {
	my ($check_path) = @_;
	my ($tmp);

	# is a symlink?
	while (-l $check_path) {
		# yes, symlink. replace with its referent.
		($tmp) = readlink $check_path;
		&log_msg(2, "Removing symbolism: '$check_path'");
		if ($check_path =~ /\//) {
			# Replace symlink name with whatever it points to.
			$check_path =~ s/\/[^\/]+$/\/$tmp/;
		} else {
			# Path with no '/' separators?
			# Probably an error...
			$check_path = $tmp;
		}
		&log_msg(2, "-->'$check_path'\n");
	}

	return $check_path;
}


# =============================================================================
# Convert $check_path to an absolute path_name.
# =============================================================================
sub make_absolute_path() {
	my ($check_path) = @_;
	my ($pwd, $old_path, $replaced, $tmp);

	# is a symlink?

	# gets absolute path to the current directory for me.
	$pwd = `pwd`;
	chomp $pwd;

	# handle relative pathnames.
	if ($check_path =~ /^\./) {
		# There are some nice perl modules that would reduce this to a single function call;
		# Argh. Can't use them without forcing user to install them first... ouch. pain. hurt.
		$old_path = $check_path;
		$check_path = $pwd . "/" . $check_path;
		&log_msg(2, "Removing relativism: '$old_path'-->'$check_path'\n");
	}

	# now try stripping down to minimal.
	$replaced=1;
	while (defined($replaced) && $replaced>0) {
		$replaced=0;

		# remove doubled /'s
		$old_path = $check_path;
		($tmp) = ($check_path =~ s/\/{2,}/\//g);
		$replaced+=$tmp;
		if ($tmp>0) {
			&log_msg(2, "Removing slashiness: '$old_path'-->'$check_path'\n");
		}

		# remove whatever/../
		$old_path = $check_path;
		($tmp) = ($check_path =~ s/[^\/]+\/\.\./\//);
		$replaced+=$tmp;
		if ($tmp>0) {
			&log_msg(2, "Removing dotism: '$old_path'-->'$check_path'\n");
		}

		# remove /./
		$old_path = $check_path;
		($tmp) = ($check_path =~ s/\/\.\//\//);
		$replaced+=$tmp;
		if ($tmp>0) {
			&log_msg(2, "Removing dotism: '$old_path'-->'$check_path'\n");
		}
	}

	return $check_path;
}


# =============================================================================
# Try to guess where the ticket kernel tree is .
# =============================================================================
sub get_ticket_path() {
	my ($check_path, $old_path);
	
	&log_msg(1, "Verbosity level: $verbosity\n");

	# my call path
	$check_path = &make_linkless_path($0);
	$check_path = &make_absolute_path($check_path);
	$old_path = $check_path;
	$check_path =~ s/\/[^\/]+$/\/..\/kernel/;
	&log_msg(2, "Guessing kernel directory: '$old_path'-->'$check_path'\n");
	$check_path = &make_absolute_path($check_path);
	&log_msg(2, "Guessing absolute directory: '$check_path'\n");

	if (!(-e $check_path) || !(-d $check_path)) {
		&log_msg(0, "My best guess as to where the TICKET kernel source would be is wrong;\n");
		&log_msg(0, "     '$check_path' does not exist, or is not a directory!\n");
		return "";
	}
	&log_msg(2, "Returning check_path '$check_path'\n");
	return $check_path;
}


# =============================================================================
# Get user to approve an action.
# =============================================================================
sub prompt_yn() {
	my ($message, $default)=@_;
	my @affirmatives = ("^yes\$", "^y\$", "^sure", "^ok\$", "^okay\$", "^okie-dokie\$", "^yup\$");
	my @negatives = ("^no\$", "^n\$", "^nope\$", "^no way\$");
	my ($idx, $yesno);

	# sensible defaults?
	if (!defined($message) || $message eq "") {
		$message = "Is this okay (y/n)? ";
	}

	# Validate response loop.
	while (1) {
		&log_msg(0, $message);
		$yesno = <STDIN>;

		if ($yesno =~ /^$/ && defined($default)) {
			$yesno=$default;
		}

		# Compare response with affirmative and negative answers.
		for ($idx=0; $idx<=$#affirmatives; $idx++) {
			if ($yesno =~ /$affirmatives[$idx]/i) {
				&log_msg(0, "\n");
				return 1;
			}
		}
		for ($idx=0; $idx<=$#negatives; $idx++) {
			if ($yesno =~ /$negatives[$idx]/i) {
				&log_msg(0, "\n");
				return 0;
			}
		}

		# No match! D'oh!
		&log_msg(0, "Please answer 'yes' nor 'no'\n");
	}
}


# =============================================================================
# Using 'lines', find the next occurance of 'regexp', starting at 'idx' and
# advancing by 'offset' after each check. This allows optionally repeated
# searches either forward or backward using a single function call.
# =============================================================================
sub find_next() {
	my ($regexp, $idx, $offset) = @_;

	&log_msg(3, "find_next: '$regexp', '$idx', '$offset'\n");
	for ($idx+=$offset; $idx<=$#lines && $idx>=0; $idx+=$offset) {
		if ($lines[$idx] =~ /$regexp/) {
			return $idx;
		}
	}

	return -1;
}


# =============================================================================
# Find 'regexp' in @lines, starting to look at 'startline' and stopping looking
# at 'stopline' (inclusive for both). Insert 'replacement' text at a location
# 'offset' away from wherever 'regexp' was found. We only use the 'cur_file'
# variable to display more useful error messages.
# =============================================================================
sub insert_find() {
	my ($cur_file, $startline, $stopline, $offset, $regexp, $replacement)  = @_;
	my $idx;

	$idx = &find_next($regexp, $startline, 1);
	if ($idx<$startline || $idx>$stopline) {
		&log_msg(0, "Failed insert of text '$replacement'...\n");
		&log_msg(0, "    I couldn't find \"$regexp\".\n");
		&log_msg(0, "I was trying to put the text $offset lines from it.\n\n");
		&log_msg(0, "Failed patch of $cur_file! Your only recourse is to patch by hand\n");
		exit(1);
	}

	# ok, put $replacement lines between ($idx-1 and $idx)...
	splice @lines, $idx+$offset, 0, ($replacement);
	return 1;
}


# =============================================================================
# 'prepatch' backs up file, then reads it into @lines.
# =============================================================================
sub prepatch() {
	my ($cur_file) = @_;
	my $backup;

	# back up before starting patch.
	$backup = "$cur_file.ticket-backup";

	if (!(-e $cur_file)) {
		&log_msg(0, "Cannot backup file '$cur_file', it doesn't exist;\n");
		&log_msg(0, "This error encountered just prior to trying to patch said file.\n");
		exit(1);
	}

	# only backup if backups do not exist; want oldest possible backups;
	# without our mucking in 'em.  Sounds backward, but think about it: otherwise
	# 'uninstallation' of TICKET code doesn't actually uninstall, it just gives
	# you an older version!
	if (!(-e $backup)) {
		if ($verbosity>0) {
			system("cp -av $cur_file $backup");
		} else {
			system("cp -a $cur_file $backup");
		}
	}

	open (PATCHFILE, "<$cur_file") || die "Failed to open '$cur_file' for patching (input)\n";
	@lines = <PATCHFILE>;
	close(PATCHFILE);
}


# =============================================================================
# 'postpatch' then writes out @lines to the file again.
# =============================================================================
sub postpatch() {
	my ($cur_file) = @_;

	open (PATCHFILE, ">$cur_file") || die "Failed to open '$cur_file' for patching (input)\n";
	print PATCHFILE @lines;
	close(PATCHFILE);
}


# =============================================================================
# Patch root Makefile;
# Configure to reference TICKET code.
# =============================================================================
sub patch_0() {
	my ($cur_file) = @_;
	my ($idx);

	#Have to change root makefile to reference new ticket code:
	#   -CORE_FILES     =kernel/kernel.o mm/mm.o fs/fs.o ipc/ipc.o
	#   +CORE_FILES     =kernel/kernel.o mm/mm.o fs/fs.o ipc/ipc.o ticket/ticket.o
	#   -SUBDIRS                =kernel drivers mm fs net ipc lib
	#   +SUBDIRS                =kernel drivers mm fs net ipc lib ticket
	#   -fs lib mm ipc kernel drivers net: dummy
	#   +fs lib mm ipc kernel drivers net ticket: dummy
	#The clean 'find and delete' logic should handle removing the proper '.o' files.

	# first replacement, CORE_FILES
	$idx = &find_next("CORE_FILES[ \t]*=[ \t]*kernel/kernel.o", 0, 1);
	if ($idx<0) {
		&log_msg(0, "Could not patch root Makefile ('$cur_file') to include ticket/ticket.o in CORE_FILES variable!\n");
		&log_msg(0, "I couldn't seem to find where CORE_FILES is set.\n");
		exit(1);
	}
	splice(@lines, $idx++, 0, "# -- TICKET -->\n");
	chomp ($lines[$idx]);
	$lines[$idx] = $lines[$idx] . " ticket/ticket.o\n";
	splice(@lines, ($idx+1), 0, "# <-- TICKET --\n");

	# second replacement, SUBDIRS
	$idx = &find_next("SUBDIRS[ \t]*=kernel", 0, 1);
	if ($idx<0) {
		&log_msg(0, "Could not patch root Makefile ('$cur_file') to include ticket in SUBDIRS variable!\n");
		&log_msg(0, "I couldn't seem to find where SUBDIRS is set.\n");
		exit(1);
	}
	splice(@lines, $idx++, 0, "# -- TICKET -->\n");
	chomp ($lines[$idx]);
	$lines[$idx] = $lines[$idx] . " ticket\n";
	splice(@lines, ($idx+1), 0, "# <-- TICKET --\n");

	# third and final replacement, the actual rule.
	$idx = &find_next("fs[ \t]+lib[ \t]+mm[ \t]+ipc", 0, 1);
	if ($idx<0) {
		&log_msg(0, "Could not patch root Makefile ('$cur_file') to include ticket in rules!\n");
		&log_msg(0, "I couldn't seem to find the 'fs lib mm ipc' rule.\n");
		exit(1);
	}
	splice(@lines, $idx++, 0, "# -- TICKET -->\n");
	chomp ($lines[$idx]);
	$lines[$idx] =~ s/net:/net ticket:/;
	$lines[$idx] .= "\n";
	splice(@lines, ($idx+1), 0, "# <-- TICKET --\n");
}


# =============================================================================
# Patch kernel's 'main' file;
# Configure to call ticket code, optionally hard-code serial console.
# =============================================================================
sub patch_1() {
	my ($cur_file) = @_;
	my $message;

	# declare ticket_init just before the initial declaration of 'init'
	&insert_find(
		$cur_file,
		0,
		$#lines,
		0,
		" init *\\(",
		"// TICKET declare ticket_init for later.\nextern void ticket_init(char **, char **);\n\n");

	# call TICKET code directly instead of execve of 'init' code.
	&insert_find(
		$cur_file,
		&find_next("dup", $#lines, -1)-1,
		$#lines,
		1,
		"dup",
		"\n\t// TICKET Just call my code. Never returns.\n\tticket_init(argv_init, envp_init);\n");
	
	# Hard-code the consoles.
	&insert_find(
		$cur_file,
		0,
		$#lines,
		0,
		"!\\*line",
		"\t// TICKET Hard-code the serial and actual consoles.\n\tchecksetup(\"console=tty0\");\n\n");

	# TODO; yank this; can be done via lilo easily enough... (?)
	$message = "Want to hard-code a serial console 'console=ttyS0,9600'?  This will give you\n";
	$message .= "kernel messages over the first serial port so long as you have ``console\n";
	$message .= "on serial port'' enabled in your kernel config (under character devices).\n";
	$message .= "These messages can be seen using a modem program like 'minicom'. This won't\n";
	$message .= "hurt performance, even if you don't use the output. Saying 'Y' merely means\n";
	$message .= "that you wouldn't always have to pass this in as a kernel parameter. Saying no\n";
	$message .= "means you don't want it, or have configured it via lilo like:\n";
	$message .= "    serial=0,9600n8\n";
	$message .= "    append=\"console=tty0 console=ttyS0,9600\"\n";
	$message .= "[Yes]: ";
	if (&prompt_yn($message, "y")) {
		&insert_find(
			$cur_file,
			0,
			$#lines,
			0,
			"!\\*line",
			"\tchecksetup(\"console=ttyS0,9600\");\n\n");
	}
}


# =============================================================================
# Patch network core devices SOURCE file;
# comment out netif_rx, as it's now a function pointer to the TICKET code.
# =============================================================================
sub patch_2() {
	my ($cur_file) = @_;
	my ($idx, $jdx);

	open (PATCHFILE, "<$cur_file") || die "Failed to open '$cur_file' for patching (input)\n";
	@lines = <PATCHFILE>;
	close(PATCHFILE);

	$idx=&find_next("netif_rx\\(struct", 0, 1);
	$jdx=&find_next("NET_RX_DROP", $idx, 1)+1;
	if ($idx>=$jdx) {
		&log_msg(0, "Failed patch of $cur_file, could not delimit netif_rx to comment it out.\n");
		exit(1);
	}

	splice @lines, $idx, $jdx-$idx+1, ("// TICKET: code removed because netif_rx overridden. It's now a function pointer, too.\n\n");
}


# =============================================================================
# Patch network core devices HEADER file;
# Change netif_rx to a function pointer (extern int (*netif_rx)(struct sk_buff *skb);)
# =============================================================================
sub patch_3() {
	my ($cur_file) = @_;
	my ($idx, $jdx);

	$idx=&find_next("netif_rx\\(struct", 0, 1);
	if ($idx<=0) {
		&log_msg(0, "Failed patch of $cur_file, could not locate netif_rx declaration\n");
		exit(1);
	}

	$lines[$idx]="extern int	(*netif_rx)(struct sk_buff *skb); // TICKET Now this is a function pointer.\n";
}


# =============================================================================
# Patch Linux version file;
# alter linux_banner to reflect new type of kernel.
# =============================================================================
sub patch_4() {
	my ($cur_file) = @_;
	my ($idx);

	$idx = &find_next("\"Linux version +\"", 0, 1);
	if ($idx<0) {
		&log_msg(0, "Could not patch Linux version number, this is not fatal. Your kernel just won't\n");
		&log_msg(0, "refer to the fact that this is TICKET version '$TICKET_VERSION'\n");
	} else {
		$lines[$idx] =~ s/Linux version/Linux-based TICKET: version '$TICKET_VERSION'. Linux version/;
	}
}


# =============================================================================
# Patch exit routines;
# optionally set auto-rebooting on panics.
# =============================================================================
sub patch_5() {
	my ($cur_file) = @_;
	my $message;

	$message  = "Do you want to hard-code a 10-second panic timeout on critical failures?\n";
	$message .= "This means that instead of hard-locking the machine on a panic, it will\n";
	$message .= "instead reboot after 10 seconds. Without a serial console or system logger,\n";
	$message .= "you may lose error messages, but such automatic reboots allow you to do\n";
	$message .= "remote administration more easily [Yes]: ";
	if (&prompt_yn($message, "y")) {
		&insert_find(
			$cur_file,
			0,
			$#lines,
			-1,
			"NORET_TYPE void do_exit",
			"\nextern int panic_timeout; // TICKET\n");
		&insert_find(
			$cur_file,
			0,
			$#lines,
			3,
			"NORET_TYPE void do_exit",
			"\n\tpanic_timeout=10; // TICKET\n");
	}
}


# =============================================================================
# Patch checksum header file; this is a bug in linux 2.4.x that hasn't been
# fixed yet (it may never be...), so I need to do it here.
# It doesn't properly include things to which it refers, so just do so.
# =============================================================================
sub patch_6() {
	my $patch;

	$lines[2] .= "/* TICKET: Required for 'struct in6_addr' */\n";
	$lines[2] .= "#include <net/ipv6.h>\n\n";
	$lines[2] .= "/* TICKET: Required for 'VERIFY_WRITE' #define */\n";
	$lines[2] .= "#include <asm/uaccess.h>\n";
}


# =============================================================================
# =============================================================================
sub patch_7() {
	my $patch;

	# Insert declaration of source tree into makefile.
	# $ticket_tree: '/home/ehw/projects/TICKET/public/code/kernel' 
	# $common_tree: '/home/ehw/projects/TICKET/public/code/common/' 
	# $source_tree: '/usr/src/linux-2.4.18'
	$lines[3] .= "# From installation script...\n";
	$lines[3] .= "SOURCE_TREE = $source_tree\n";
}


# =============================================================================
# Move the given file back over the original.
# =============================================================================
sub revert_file() {
	my ($filename) = @_;
	my $backup;

	$backup = "$filename.ticket-backup";

	if (!(-e $backup)) {
		# pretend this is a 'soft' error; user probably did something foolish.
		# (like removing backup files).
		&log_msg(0, "Cannot revert backup of '$filename'.\n");
		&log_msg(0, "\t'$backup' doesn't exist\n");
	} else {
		# move the backup over the 'real' file.
		if ($verbosity>0) {
			system("mv -vf $backup $filename");
		} else {
			system("mv -f $backup $filename");
		}
	}
}


# =============================================================================
# Perform code uninstallation.
# This code just removes the directories we've copied,
# then moves the backups we've made back over the files we've altered.
# =============================================================================
sub do_uninstall() {
	my $options;
	my $file;

	if ($verbosity>0) {
		$options = "v";
	} else {
		$options = "";
	}

	# remove directories copied
	system("rm -rf$options $source_tree/ticket/");
	system("rm -rf$options $source_tree/ticket/common/");

	# 'unpatch' files
	&revert_file("$source_tree/Makefile");
	&revert_file("$source_tree/init/main.c");
	&revert_file("$source_tree/net/core/dev.c");
	&revert_file("$source_tree/include/linux/netdevice.h");
	&revert_file("$source_tree/init/version.c");
	&revert_file("$source_tree/kernel/exit.c");
	&revert_file("$source_tree/include/asm-i386/checksum.h");
}


# =============================================================================
# Perform code installation.
# =============================================================================
sub do_install() {
	my ($idx);

	# ----------------------------------------
	# Get directories that we want to work with.
	# ----------------------------------------
	if ($ticket_tree eq "") {
		$ticket_tree = &get_ticket_path();
	}
	# get where the common 
	$common_tree = &make_linkless_path("$ticket_tree/../common/");
	$common_tree = &make_absolute_path($common_tree);

	# start real installation stuff.
	if (-d "$source_tree/ticket/") {
		&log_msg(0, "ERROR: TICKET already seems to be installed into the kernel source here.\n");
		exit 1;
	}

	# prompt to see if this is correct
	if (!(&prompt_yn("I will install TICKET kernel code from '$ticket_tree' and '$common_tree' to '$source_tree'\nIs this correct [Yes]? ", "y"))) {
		&usage("Ok, then pass in parameters by hand.");
	}

	# ----------------------------------------
	# INSTALL
	# ----------------------------------------
	if ($verbosity>0) {
		&log_msg(1, "Copying TICKET kernel code...\n");
		system("cp -av $ticket_tree/ $source_tree/ticket/");
		unlink("$source_tree/ticket/common"); # 20021205: symlink in distro; real dir in install.
		system("cp -av $common_tree/ $source_tree/ticket/common/");
	} else {
		system("cp -a $ticket_tree/ $source_tree/ticket/");
		unlink("$source_tree/ticket/common"); # 20021205: symlink in distro; real dir in install.
		system("cp -a $common_tree/ $source_tree/ticket/common/");
	}

	# Then patch up a few C source and header files to connect with the TICKET code.
	# 'prepatch' backs up file, then reads it into @lines.
	# then the patch routine mucks with the data in @lines.
	# 'postpatch' then writes out @lines to the file again.
	$idx="$source_tree/Makefile"                   ; &prepatch($idx); &patch_0($idx); &postpatch($idx);
	$idx="$source_tree/init/main.c"                ; &prepatch($idx); &patch_1($idx); &postpatch($idx);
	$idx="$source_tree/net/core/dev.c"             ; &prepatch($idx); &patch_2($idx); &postpatch($idx);
	$idx="$source_tree/include/linux/netdevice.h"  ; &prepatch($idx); &patch_3($idx); &postpatch($idx);
	$idx="$source_tree/init/version.c"             ; &prepatch($idx); &patch_4($idx); &postpatch($idx);
	$idx="$source_tree/kernel/exit.c"              ; &prepatch($idx); &patch_5($idx); &postpatch($idx);
	$idx="$source_tree/include/asm-i386/checksum.h"; &prepatch($idx); &patch_6($idx); &postpatch($idx);
	$idx="$source_tree/ticket/Makefile"            ; &prepatch($idx); &patch_7($idx); &postpatch($idx);
}


# =============================================================================
# Code entry point.
# Parse command line, then either install or uninstall code.
# =============================================================================
sub main() {
	if (!open(LOGFILE, ">./install.log")) {
		print "Failed to open ./install.log; read-only filesystem?\n";
		print "Logging to /tmp/TICKET_install.log instead.\n";
		open(LOGFILE, ">/tmp/TICKET_install.log") || die "ERROR: Failed to open /tmp/TICKET_install.log!\n";
	}

	&parse_cmd_line();

	if ($mode eq 'i')  {
		&do_install();
	} else {
		&do_uninstall();
	}
}

# Yes, I am a C programmer. How can you tell?
&main();

