/*
# TICKET - Traffic Information Collection Kernel with Exact Timing
#
# This file is part of TICKET, the Traffic Information Collection Kernel with
# Exact Timing. This program and all associated files are Copyright (C) 2001
# by Eric H. Weigle of Los Alamos National Laboratory for the Regents of the
# University of California.
# 
# Version:	2.0.0-alpha	2003/02/20
# 
# Authors: Eric Weigle, ehw@lanl.gov or radiant-software@lanl.gov
# 
# History:
# 
#   2002-12-24 1.0.0-a  initial release
#   2003-02-20 2.0.0-a  first moderately usable release
# 
# License:
# 
# Copyright 2002. The Regents of the University of California.
# 
# This material was produced under U.S. Government contract
# W-7405-ENG-36 for Los Alamos National Laboratory, which is operated
# by the University of California for the U.S. Department of Energy.
# The Government is granted for itself and other acting on its behalf
# a paid-up, non-exclusive, irrevocable worldwide license in the
# material to reproduce, prepare derivative works, and perform
# publicly and display publicly. Beginning five (5) years after
# November 6, 2002, subject to additional five-year worldwide
# renewals, the Government is granted for itself and others acting on
# its behalf a paid-up, nonexclusive, irrevocable worldwide license
# in this material to reproduce, prepare derivative works, distribute
# copies to the public, perform publicly and display publicly, and to
# permit others to do so. NEITHER THE UNITED STATES NOR THE UNITED
# STATES DEPARTMENT OF ENERGY, NOR THE UNIVERSITY OF CALIFORNIA, NOR
# ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR
# ASSUMES ANY LEGAL LIABLITY OR RESPONSIBILITY FOR THE ACCURACY,
# COMPLETENESS, OR USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT,
# OR PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE
# PRIVATELY OWNED RIGHTS.
# 
# Additionally, this program is free software; you can distribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2 of
# the License, or any later version. Accordingly, this program is
# distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details. You should have received a copy of the
# GNU General Public License along with this program; if not, write
# to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# ------------------------------------------------------------
# 
# You may contact the author (Eric H. Weigle) via email at ehw@lanl.gov,
# or via postal mail at MS D451, P.O. Box 1663, Los Alamos NM  87545.
# 
# Should you fail to reach the author at these addresses, please refer to the
# following websites for updated information or other means by which he can
# be contacted:
# 	http://public.lanl.gov/ehw/ 
# 	http://public.lanl.gov/radiant/
#
*/

// #############################################################################
// Writer program. Saves data coming from kernel box to local disk.
// #############################################################################
#include "../common/network.h"
#include "../common/dump.h"
#include "../common/misc.h"
#include "../common/config.h"
#include "../common/doubly.h"
#include <signal.h>           // sigaction
#include <time.h>             // time.
#include <sys/stat.h>         // for mkdir
#include <sys/types.h>        // for mkdir, open()
#include <unistd.h>           // close(), sleep
#include <stdlib.h>           // malloc()
#include <sys/vfs.h>          // statfs

#define MAX_OUT_OF_ORDER 127

// =============================================================================
// GLOBAL VARIABLES
// =============================================================================
// The C[++] standard may define these to be nulled out,
// but I like to do so myself.
unsigned long long packet_count=0;
unsigned long long packet_drops=0;
unsigned long cpu_speed, trace_start;
int sockfd=0, verbosity=0;
int bufsize=0;
// For writing to files.
FILE *DUMPFILE=NULL;
char **devices=NULL;
int dev_count;
// Scratch buffer used in various places.
int tmpbuf_size=0;
char *tmpbuf=NULL;

// =============================================================================
// LOCAL FUNCTIONS, declared
// =============================================================================
// Tell the user how they screwed up ;)
void usage(char *error);
// verify validity of command-line parameters
int check_devices(void);
// parse command-line options
void parse_command_line(int argc, char *argv[]);
// Return the amount of space free on 'device_name'
int disk_free(char *device_name);
// Function called on ^C or other deadly event
void terminate(int sig);
// split the dump file when it gets to the flag size.
void dump_file_split(void);
// list of dump packets and memory we're working with.
struct dlinked_list packet_list;
struct dlinked_list dirty_list;


// =============================================================================
// how to call program.
// =============================================================================
void usage(char *error) {
	if (error) { printf("\nUSAGE ERROR: %s\n", error); }
	printf("USAGE: ./writer <-v+> <-b=bufsize> [-c=cpu_speed] [-s=trace_start] [directory]+\n\n");
	printf("    -v, -vv, -vvv: run verbosely, very verbosly, and in debugging mode\n");
	printf("    -b=bufsize: set the socket buffer size to this value.\n"); 
	printf("    -c=cpu_speed: the collection machine's cpu speed.\n");
	printf("        required to convert rdtsc() timestamps to struct timevals.\n");
	printf("    -s=trace_start: seconds since the epoch (unix time() function)\n");
	printf("        required to convert rdtsc() timestamps to struct timevals.\n");
	printf("WARNING: if cpu_speed and trace_start are NOT passed in, the program assumes\n");
	printf("    them to be the same as on localhost! This is probably fine for the start\n");
	printf("    time but it's probably WRONG for the cpu speed!\n\n");
	printf("Pass in any number of directories in which to store output files;\n");
	printf("we write files to each directory until its device is full, then move on.\n");
	exit(1);
}


// =============================================================================
// Check that each passed-in parameter is actually on a separate device.
// If two or more are on the same device, the striping code fails (we
// muck with the first one until its full, then when we hit the second
// one the disk is already full).
// =============================================================================
int check_devices(void) {
	struct stat *file_info;
	int idx, jdx;

	// check that we have some data.
	if (dev_count<1) {
		printf("ERROR: Too few output directories on command line\n");
		return 0;
	}

	// allocate memory for stat object array (one per device).
	if ((file_info=(struct stat *)malloc(dev_count*sizeof(struct stat)))==NULL) {
		printf("Error: Failed to allocate file_info structure!\n");
		return 0;
	}

	// first need to get info for all of 'em.
	for (tmpbuf_size=idx=0; idx<dev_count; idx++) {
		if (verbosity>1) {
			printf("Checking file at '%s'...\n", devices[idx]);
		}
		// this code ensures that the temporary buffer is large enough to hold
		// what we'll use it for.
		if (strlen(devices[idx])>tmpbuf_size) {
			tmpbuf_size = strlen(devices[idx]);
		}

		if (stat(devices[idx], &(file_info[idx]))!=0) {
			printf("Error: Failed to stat FILE '%s' because '%s'.\n", devices[idx], strerror(errno));
			return 0;
		}
	}

	// figure out 2^N size for temp. buffer (Include space for max pathname).
	idx=tmpbuf_size+32;
	for (tmpbuf_size=1; tmpbuf_size<idx; tmpbuf_size *=2) {
		;
	}
	// create temporary buffer
	if ((tmpbuf=(char *)malloc(tmpbuf_size))==NULL) {
		printf("Could not create scratch buffer!\n");
		return 0;
	}

	// compare all devices to ensure each is unique. (O(n^2) comparison....)
	for (idx=0; idx<(dev_count-1); idx++) {
		for (jdx=(idx+1); jdx<dev_count; jdx++) {
			if (memcmp((void *)&(file_info[idx].st_dev), (void *)&(file_info[jdx].st_dev), sizeof(dev_t))==0) {
				printf("Error: '%s' and '%s' are on same device.\n", devices[idx], devices[jdx]);
				printf("I can't figure out striping in this case\n");
				return 0;
			}
		}
	}
	if (verbosity>2) {
		printf("\tDevices look good; each is unique.\n");
	}

	// free memory
	free(file_info);

	// done!
	return 1;
}


// =============================================================================
// get command line parameters
// =============================================================================
void parse_command_line(int argc, char *argv[]) {
	int idx, len;

	// initialize junk.
	bufsize=0;
	cpu_speed=0;
	trace_start=0;
	verbosity=0;

	// Allocate space for maximum number of filenames
	dev_count=0;
	if ((devices = (char **)calloc(argc-1,sizeof(char *)))==NULL) {
		perror("Failed to allocate file name pointers!\n");
		exit(1);
	}

	for (idx=1; idx<argc; idx++) {
		if (argv[idx][0]=='-') {
			/* is it a flag? */
			switch(argv[idx][1]) {
				case 'b': case 'B': /* buffer size */
					if (argv[idx][2]=='=') {
						argv[idx][2]=' ';
					}
					if (sscanf(&(argv[idx][2]), "%d", &bufsize)!=1) {
						usage("Failed scan of bufsize string!");
					}
					if (bufsize<=0) {
						usage("Scanned bad (negative or zero) bufsize string!");
					}
					continue;
				case 'c': case 'C': /* cpu speed */
					cpu_speed = atol(&(argv[idx][3]));
					continue;
				case 's': case 'S': /* trace start */
					trace_start = atol(&(argv[idx][3]));
					continue;
				case 'v': case 'V':  /* verbosity */
					verbosity=strlen(argv[idx])-1;
					continue;
				default:
					usage("Bad option on command line!");
					continue;
			}
		} else {
			/* not a flag, must be a device directory */
			// Not an option. Add filename to list.
			len = strlen(argv[idx]) + 1;
			if ((devices[dev_count] = (char *)malloc(len*sizeof(char)))==NULL) {
				perror("Failed to allocate file name pointers!\n");
				exit(1);
			}
			memcpy(devices[dev_count], argv[idx], len);
			dev_count++;
		}
	}

	// check input here.
	// buffer size and verbosity are optional.
	if (cpu_speed<=0) {
		// This may be a very bad assumption to make (???)
		cpu_speed = get_cpuinfo();
		if (cpu_speed<=0) {
			usage("Please pass in CPU speed on command line; I couldn't figure it out\n");
		}
		if (verbosity>0) {
			printf("WARNING: You didn't enter the CPU speed of the collection machine!\n");
			printf("    Assuming it's the same as localhost... %lu Hz\n", cpu_speed);
		}
	}
	if (trace_start<=0) {
		// This is probably an OK assumption.
		trace_start = time(NULL);
		if (verbosity>0) {
			printf("WARNING: You didn't enter the trace start time!\n"); 
			printf("    Assuming localhost's clock is correct... %lu seconds since epoch.\n", trace_start);
		}
	}
	if (!check_devices()) {
		usage("Device parameter(s) passed in are invalid!");
	}

	// Ok, give some info now.
	if (verbosity>0) {
		int dev_idx;
		printf("CPU speed: %lu Hertz.\n", cpu_speed);
		printf("Trace start (now): %lu seconds.\n", trace_start);
		printf("File size will be %d bytes, taking up about %d blocks/file.\n", FILE_SIZE_BYTES, FILE_SIZE_BLOCKS);
		for (dev_idx=0; devices[dev_idx]!=NULL; dev_idx++) {
			printf("\t'%s' (device %d) will hold about %d files.\n",
					devices[dev_idx], dev_idx, disk_free(devices[dev_idx]));
		}
		// have to reset before continuing!
		dev_idx=0;
	}
}



// =============================================================================
// How much disk space (in terms of FILES) is left on the given device?
// See statfs(2) for some useful info.
//
// The arithmetic below uses integers (not long longs or doubles).
// On a 32-bit system it should work for disks up to 2TB in size.
//
// This algorithm has some problems, such as the fact that FILE_SIZE_BLOCKS
// is not really known beforehand. To account for this I add some fluff to
// the factors (+4k fluff in the FILE_SIZE_BLOCKS #define).
// =============================================================================
int disk_free(char *device_name) {
	struct statfs fs_info;

	// now get filesystem information.
	if (verbosity>2) {
		printf("Checking filesystem at '%s'\n", device_name);
	}
	if (statfs(device_name, &fs_info)!=0) {
		printf("Failed to stat '%s' because '%s'\n", device_name, strerror(errno));
		exit(1); // just return zero and 'soft fail'?
	}

	// root reserves some space, return appropriate value depending on whether we're root
	// or a user who cannot access that disk space.
	if (getuid()==0) {
		// printf("Root availability... %ld blocks, %ld files\n", fs_info.f_bsize * (fs_info.f_bfree/1024), (fs_info.f_bsize * (fs_info.f_bfree/1024))/FILE_SIZE_BLOCKS ); // 1KB blocks available to users
		return((fs_info.f_bsize * (fs_info.f_bfree/1024))/FILE_SIZE_BLOCKS - 1);
	} else {
		// printf("User availability... %ld blocks, %ld files\n", fs_info.f_bsize * (fs_info.f_bavail/1024), (fs_info.f_bsize * (fs_info.f_bavail/1024))/FILE_SIZE_BLOCKS ); // 1KB blocks available to users
		return((fs_info.f_bsize * (fs_info.f_bavail/1024))/FILE_SIZE_BLOCKS - 1);
	}

	return 0;
}


// =============================================================================
// server's shutdown routine, run asynchronously.
// =============================================================================
void terminate(int sig) {
	int idx;
	dump_packet *packet;

	if (sockfd!=0) { close(sockfd); }
	if (tmpbuf!=NULL) { free(tmpbuf); }

	if (devices!=NULL) {
		for (idx=0; devices[idx]!=NULL; idx++) {
			free(devices[idx]);
		}
		free(devices);
	}

	// free lists.
	while (dirty_list.count>0) {
		dlist_pop_head(&dirty_list, (void **)&packet);
		free(packet);
	}

	while (packet_list.count>0) {
		dlist_pop_head(&packet_list, (void **)&packet);
		free(packet);
		packet_drops++;
	}

	printf("Received %llu dump packets, lost %llu dump packets.\n", packet_count, packet_drops);

	exit(0);

}

// =============================================================================
// Write the TCPDUMP file header.
// =============================================================================
#ifdef DUMP_MODE_TCPDUMP
struct tcpdump_file_header {
	unsigned int magic;
	unsigned short version_major;
	unsigned short version_minor;
	int thiszone;
	unsigned int sigfigs;
	unsigned int snaplen;
	unsigned int linktype;
};

int write_tcpdump_header(void) {
	struct tcpdump_file_header hdr;

	hdr.magic         =  0xa1b2c3d4; /* little-endian/big-endian fixer */
	hdr.version_major =  2; /* pcap v. 2.4 file format */
	hdr.version_minor =  4;
	hdr.thiszone      =  0;  /* gmt to local correction */
	hdr.sigfigs       =  0;  /* accuracy of timestamps */
	hdr.snaplen       =  OUT_SNAPLEN; /* max length saved portion of each pkt; 68 or 96 */
	hdr.linktype      =  1;  /* data link type (LINKTYPE_*); DLT_EN10MB=1 */

	return(fwrite((char *)&hdr, sizeof(hdr), 1, DUMPFILE) == 1 ? 1 : 0);
}
#endif

// =============================================================================
// split the dump file when it gets to the flag size.
// =============================================================================
void dump_file_split(void) {
	static int abs_file_num=0;
	static int rel_file_num=0;
	static int dev_idx=0;

	// tell user we're splitting.
	if (verbosity>0) {
		printf("S(%d) ", abs_file_num);
	}

	// close last file if it exists.
	if (DUMPFILE!=NULL && fclose(DUMPFILE)!=0) {
		perror("Failure closing previous dumpfile!\n");
		exit(-1);
	}

	// Need to move to new mount point/device
	// Add some fluff here (2 instead of 1):
	//   this ensures we always leave space for an extra file per device.
	while (disk_free(devices[dev_idx])<2) {
		if (devices[dev_idx]==NULL || devices[dev_idx+1]==NULL) {
			printf("I've written %d files, filling all given devices. Terminating\n", abs_file_num);
			terminate(0);
		}
		dev_idx++;

		// Make the directory. see chmod(2) for permission info
		// TODO: check lengths & stuff.
		snprintf(tmpbuf, tmpbuf_size, "%s/%03d/", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR);
		if (mkdir(tmpbuf, S_IRWXU|S_IRWXG|S_IRWXO)<0) {
			if (errno != EEXIST) {
				printf("Could not make directory '%s' because '%s'\n", tmpbuf, strerror(errno));
				terminate(1);
			}
		}

		rel_file_num=0;
	}

	// move to next directory when this one full.
	if (!(abs_file_num%DUMP_FILES_PER_DIR)) {
		// Make the directory. see chmod(2) for permission info
		// TODO: check lengths & stuff.
		snprintf(tmpbuf, tmpbuf_size, "%s/%03d/", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR);
		if (mkdir(tmpbuf, S_IRWXU|S_IRWXG|S_IRWXO)<0) {
			if (errno != EEXIST) {
				printf("Could not make directory '%s' because '%s'\n", tmpbuf, strerror(errno));
				terminate(1);
			}
		}

		rel_file_num=0;
	}

	// open new file.
	// TODO: check lengths & stuff.
	snprintf(tmpbuf, tmpbuf_size, "%s/%03d/file.%06d", devices[dev_idx], abs_file_num/DUMP_FILES_PER_DIR, abs_file_num);
	if ((DUMPFILE=fopen(tmpbuf, "w"))==NULL) {
		printf("Can't open dumpfile '%s'\n", tmpbuf);
		perror("Error was");
		exit(-1);
	}

#ifdef DUMP_MODE_TCPDUMP
	// if dumping in mode TCPDUMP, must write file header.
	write_tcpdump_header();
#endif

	// increment file counter.
	abs_file_num++;
	rel_file_num++;
}


// =============================================================================
// Save system information including the time and CPU speed for post analysis.
// =============================================================================
void save_system_info(void) {
	FILE *status_file;
	
	// TODO: check lengths & stuff.
	snprintf(tmpbuf, tmpbuf_size, "%s/system_info.txt", devices[0]);

	if ((status_file=fopen(tmpbuf, "w"))==NULL) {
		printf("WARNING! Failed to open system info file '%s'\n", tmpbuf);
		return;
	}
	fprintf(status_file, "The UNIX time is %lu.\n\n", time(NULL));
	fprintf(status_file, "Information from /proc/cpuinfo follows:\n");
	fclose(status_file);

	// TODO: check lengths & stuff.
	snprintf(tmpbuf, tmpbuf_size, "cat /proc/cpuinfo >> %s/system_info.txt", devices[0]);
	system(tmpbuf);
}


// =============================================================================
// Compare packet sequence numbers; for sorting.
// =============================================================================
static int pkt_seqno_compare(void *pkt_a, void *pkt_b) {
	return (((dump_packet *)pkt_b)->seqno - ((dump_packet *)pkt_a)->seqno);
}


// =============================================================================
// Save the packet to the current file.
// =============================================================================
#ifdef DUMP_MODE_TICKET
int save_dump_packet(dump_packet *cur_pkt) {
	printf("Got packet %llu, saving data to file.\n", cur_pkt->seqno);
	return(fwrite(cur_pkt->data, sizeof(collected_headers), DUMPS_PER_PACKET, DUMPFILE)==DUMPS_PER_PACKET);
}
#endif
#ifdef DUMP_MODE_TCPDUMP
struct tcpdump_header {
	int seconds;
	int microseconds;
	unsigned int caplen;
	unsigned int pktlen;
};

int save_dump_packet(dump_packet *cur_pkt) {
	int idx;
	struct tcpdump_header hdr;
	collected_headers *cur_hdr;

	printf("Got packet %llu, saving data to file.\n", cur_pkt->seqno);

	// convert each packet and save to file.
	for (idx=0; idx<DUMPS_PER_PACKET; idx++) {
		cur_hdr = &(cur_pkt->data[idx]);

		hdr.seconds = (cur_hdr->timestamp/cpu_speed)+trace_start;
		hdr.microseconds = (((double)(cur_hdr->timestamp%cpu_speed))/cpu_speed)*1000000;
		hdr.caplen = (cur_hdr->pktlen<=OUT_SNAPLEN ? cur_hdr->pktlen : OUT_SNAPLEN);
		hdr.pktlen = cur_hdr->pktlen;

		if (fwrite(&hdr, sizeof(struct tcpdump_header), 1, DUMPFILE)!=1) {
			printf("Failed header write on packet %d\n", idx);
			return 0;
		}
		if (fwrite((char *)cur_hdr->snap, hdr.caplen, 1, DUMPFILE)!=1) {
			printf("Failed data write on packet %d\n", idx);
			return 0;
		}
	}
	return 1;
}
#endif

// =============================================================================
// The main routine
// =============================================================================
int main(int argc, char *argv[]) {
	struct sigaction my_sigact;
	dump_packet *cur_pkt;
	struct sockaddr_in serv_addr;
	int len, addr_len;
	unsigned long long next_seqno=0;

	if (sizeof(struct dump_packet_ticket)==sizeof(struct dump_packet_tcpdump)) {
		printf("Warning: Can't detect mode by size of packets... ");
#ifdef DUMP_MODE_TICKET
		printf("Using TICKET mode.\n");
#endif
#ifdef DUMP_MODE_TCPDUMP
		printf("Using TCPDUMP mode.\n");
#endif
	}

	// set up list junk.
	memset(&packet_list, '\0', sizeof(struct dlinked_list));
	memset(&dirty_list, '\0', sizeof(struct dlinked_list));
	packet_list.compfn = &pkt_seqno_compare;
	dirty_list.compfn = &pkt_seqno_compare;

	// fill dirty list. Need one more packet for currently-processed one.
	while (dirty_list.count<(MAX_OUT_OF_ORDER+1)) {
		if ((cur_pkt=(dump_packet *)malloc(sizeof(dump_packet)))==NULL) {
			printf("Couldn't create dump packet input ring index %d!\n", dirty_list.count);
			terminate(1);
		}
		dlist_push_head(&dirty_list, (void *)cur_pkt);
	}

	// parse command line.
	parse_command_line(argc, argv);

	// Handle termination cleanly.
	bzero(&my_sigact, sizeof(struct sigaction));
	my_sigact.sa_handler = &(terminate);
	if (sigaction(SIGINT, &my_sigact, NULL)) {
		perror("Could not set up signal handlers.\n");
		return 1;
	}

	// network startup
	if (!(sockfd=create_udp_socket(NULL, &serv_addr, bufsize, verbosity))) {
		printf("Failed to create a socket!\n");
		exit(1);
	}

	// save cpu speed & time.
	save_system_info();

	if (verbosity>0) { printf("Ready to receive dump packets...\n"); }

	// now loop forever for messages from client
	while (1) {
		// get a packet to work with.
		dlist_pop_head(&dirty_list, (void **)&cur_pkt);

		// read a packet
		addr_len = sizeof(serv_addr);
		if ((len=recvfrom(sockfd, cur_pkt, sizeof(dump_packet), MSG_WAITALL,
				(struct sockaddr *) &serv_addr, (socklen_t *) &addr_len))
				!= sizeof(dump_packet)) {
			printf("Error '%s'; Received mis-sized packet (length %d, not %d)!\n", strerror(errno), len, sizeof(dump_packet));
			if (len==sizeof(struct dump_packet_ticket)) { 
				printf("But it is the right size if we had been in TICKET mode. You need to recompile!\n");
			}
			if (len==sizeof(struct dump_packet_tcpdump)) {
				printf("But it is the right size if we had been in TCPDUMP mode. You need to recompile!\n");
			}
			goto packet_error;
		}

		// tell user
		if (verbosity>0) { printf("."); }

		// --------------------------------------------
		// Good (hopefully common) case first.
		// --------------------------------------------
		while (cur_pkt && cur_pkt->seqno==next_seqno) {
			// split file if required.
			if ((packet_count%RECORDS_PER_FILE)==0) {
				dump_file_split();
			}

			// save to file.
			if (!save_dump_packet(cur_pkt)) {
				printf("Failed write to file!!! Error '%s'\n", strerror(errno));
				goto packet_error;
			}

			packet_count++;
			next_seqno++;
			dlist_push_head(&dirty_list, (void *)cur_pkt);

			// tell user we're writing.
			if (verbosity>0) { printf("W"); }
	
			// returns NULL if list empty.
			dlist_pop_head(&packet_list, (void **)&cur_pkt);
		}

		// --------------------------------------------
		// Error case next.
		// --------------------------------------------
		if (cur_pkt) {
			// Drop packets that are duplicated or so badly out of sequence that
			// we've already declared them LOST.
			if (cur_pkt->seqno < next_seqno) {
				printf("Got prior packet... may be a duplicate or already marked as lost.\n");
				printf("Its sequence number, %llu, is BEFORE the current one, %llu\n", cur_pkt->seqno, next_seqno);
				goto packet_error;
			}

			// Save out-of-order packets for next time.
			printf("Out-of-order packet, wanted %llu got %llu\n", next_seqno, cur_pkt->seqno);
			dlist_insert(&packet_list, (void *)cur_pkt);

			// Handle loss. This is why we needed to allocate that extra packet above;
			// on next iteration it becomes cur_packet and all packets allocated are used.
			if (packet_list.count>=MAX_OUT_OF_ORDER) {
				dlist_pop_head(&packet_list, (void **)&cur_pkt);
				printf("Packet(s) lost; %llu to %llu...\n", next_seqno, cur_pkt->seqno-1);
				// if re-starting writer after death; will get massive loss count.
				// I _think_ this is the right way to count it.
				packet_drops += cur_pkt->seqno - next_seqno;
				next_seqno = cur_pkt->seqno;
				dlist_push_head(&packet_list, (void *)cur_pkt);
			}
		}

		continue;
packet_error:
		dlist_push_head(&dirty_list, (void *)cur_pkt);
		packet_drops++;
	}

	terminate(0);
}

